from .basic_octree import *
from .contour_finding import *
from .depth_first_octree import *
from .fortran_reader import *
from .grid_traversal import *
from .image_utilities import *
from .interpolators import *
from .line_integral_convolution import *
from .marching_cubes import *
from .mesh_utilities import *
from .misc_utilities import *
from .particle_mesh_operations import *
from .points_in_volume import *
from .quad_tree import *
from .write_array import *



import contextlib
import os
import shutil
import sys
import tempfile
import unittest
import unittest.mock as mock
from io import StringIO

import yt.config
import yt.utilities.command_line
from yt.utilities.configure import YTConfig

_TEST_PLUGIN = "_test_plugin.py"
# NOTE: the normalization of the crazy camel-case will be checked
_DUMMY_CFG_INI = f"""[yt]
logLevel = 49
pluginfilename = {_TEST_PLUGIN}
boolean_stuff = True
chunk_size = 3
"""

_DUMMY_CFG_TOML = f"""[yt]
log_level = 49
plugin_filename = "{_TEST_PLUGIN}"
boolean_stuff = true
chunk_size = 3
"""


@contextlib.contextmanager
def captureOutput():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


class SysExitException(Exception):
    pass


class TestYTConfig(unittest.TestCase):
    def setUp(self):
        self.xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
        self.tmpdir = tempfile.mkdtemp()
        os.environ["XDG_CONFIG_HOME"] = self.tmpdir
        os.mkdir(os.path.join(self.tmpdir, "yt"))

        # run inside another temporary directory to avoid poluting the
        # local space when we dump configuration to a local yt.toml file
        self.origin = os.getcwd()
        os.chdir(tempfile.mkdtemp())

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        if self.xdg_config_home:
            os.environ["XDG_CONFIG_HOME"] = self.xdg_config_home
        else:
            os.environ.pop("XDG_CONFIG_HOME")
        os.chdir(self.origin)

    def _runYTConfig(self, args):
        args = ["yt", "config"] + args
        retcode = 0

        with mock.patch.object(sys, "argv", args), mock.patch(
            "sys.exit", side_effect=SysExitException
        ) as exit, captureOutput() as output:
            try:
                yt.utilities.command_line.run_main()
            except SysExitException:
                args = exit.mock_calls[0][1]
                retcode = args[0] if len(args) else 0
        return {"rc": retcode, "stdout": output[0], "stderr": output[1]}

    def _testKeyValue(self, key, val_set, val_get):
        info = self._runYTConfig(["set", "yt", key, str(val_set)])
        self.assertEqual(info["rc"], 0)

        info = self._runYTConfig(["get", "yt", key])
        self.assertEqual(info["rc"], 0)
        self.assertEqual(info["stdout"].strip(), str(val_get))

        info = self._runYTConfig(["rm", "yt", key])
        self.assertEqual(info["rc"], 0)

    def _testKeyTypeError(self, key, val1, val2, expect_error):
        info = self._runYTConfig(["set", "yt", key, str(val1)])
        self.assertEqual(info["rc"], 0)

        if expect_error:
            with self.assertRaises(TypeError):
                info = self._runYTConfig(["set", "yt", key, str(val2)])
        else:
            info = self._runYTConfig(["set", "yt", key, str(val2)])

        info = self._runYTConfig(["rm", "yt", key])
        self.assertEqual(info["rc"], 0)


class TestYTConfigCommands(TestYTConfig):
    def testConfigCommands(self):
        def remove_spaces_and_breaks(s):
            return "".join(s.split())

        self.assertFalse(os.path.exists(YTConfig.get_global_config_file()))

        info = self._runYTConfig(["--help"])
        self.assertEqual(info["rc"], 0)
        self.assertEqual(info["stderr"], "")
        self.assertIn(
            remove_spaces_and_breaks("Get and set configuration values for yt"),
            remove_spaces_and_breaks(info["stdout"]),
        )

        info = self._runYTConfig(["list"])
        self.assertEqual(info["rc"], 0)
        self.assertEqual(info["stdout"], "")

        self._testKeyValue("internals.parallel", True, True)
        self._testKeyValue(
            "test_data_dir", "~/yt-data", os.path.expanduser("~/yt-data")
        )
        self._testKeyValue(
            "test_data_dir", "$HOME/yt-data", os.path.expandvars("$HOME/yt-data")
        )

        with self.assertRaises(KeyError):
            self._runYTConfig(["get", "yt", "foo"])

        # Check TypeErrors are raised when changing the type of an entry
        self._testKeyTypeError("foo.bar", "test", 10, expect_error=True)
        self._testKeyTypeError("foo.bar", "test", False, expect_error=True)

        # Check no type error are raised when *not* changing the type
        self._testKeyTypeError("foo.bar", 10, 20, expect_error=False)
        self._testKeyTypeError("foo.bar", "foo", "bar", expect_error=False)

    def tearDown(self):
        if os.path.exists(YTConfig.get_global_config_file()):
            os.remove(YTConfig.get_global_config_file())
        super().tearDown()


class TestYTConfigGlobalLocal(TestYTConfig):
    def setUp(self):
        super().setUp()
        with open(YTConfig.get_local_config_file(), mode="w") as f:
            f.writelines("[yt]\n")
        with open(YTConfig.get_global_config_file(), mode="w") as f:
            f.writelines("[yt]\n")

    def testAmbiguousConfig(self):
        info = self._runYTConfig(["list"])
        self.assertFalse(len(info["rc"]) == 0)

        for cmd in (["list", "--local"], ["list", "--global"]):
            info = self._runYTConfig(cmd)
            self.assertEqual(info["rc"], 0)



import os
import weakref
from collections import defaultdict
from itertools import product
from pathlib import Path
from typing import Optional, Tuple

import numpy as np

from yt.arraytypes import blankRecordArray
from yt.data_objects.index_subobjects.octree_subset import OctreeSubset
from yt.data_objects.particle_filters import add_particle_filter
from yt.data_objects.static_output import Dataset
from yt.funcs import mylog, setdefaultattr
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.oct_container import RAMSESOctreeContainer
from yt.geometry.oct_geometry_handler import OctreeIndex
from yt.utilities.cython_fortran_utils import FortranFile as fpu
from yt.utilities.lib.cosmology_time import friedman
from yt.utilities.on_demand_imports import _f90nml as f90nml
from yt.utilities.physical_constants import kb, mp
from yt.utilities.physical_ratios import cm_per_mpc

from .definitions import (
    OUTPUT_DIR_EXP,
    OUTPUT_DIR_RE,
    STANDARD_FILE_RE,
    field_aliases,
    particle_families,
    ramses_header,
)
from .field_handlers import get_field_handlers
from .fields import _X, RAMSESFieldInfo
from .hilbert import get_cpu_list
from .io_utils import fill_hydro, read_amr
from .particle_handlers import get_particle_handlers


class RAMSESFileSanitizer:
    """A class to handle the different files that can be passed and associated
    safely to a RAMSES output."""

    root_folder = None  # Path | None: path to the root folder
    info_fname = None  # Path | None: path to the info file
    group_name = None  # str | None: name of the first group folder (if any)

    def __init__(self, filename):

        # Make the resolve optional, so that it works with symlinks
        paths_to_try = (Path(filename), Path(filename).resolve())

        self.original_filename = filename

        self.output_dir = None
        self.info_fname = None

        check_functions = (self.test_with_standard_file, self.test_with_folder_name)

        # Loop on both the functions and the tested paths
        for path, check_fun in product(paths_to_try, check_functions):
            ok, output_dir, group_dir, info_fname = check_fun(path)
            if ok:
                break

        # Early exit if the ok flag is False
        if not ok:
            return

        self.root_folder = output_dir
        self.group_name = group_dir.name if group_dir else None
        self.info_fname = info_fname

    def validate(self) -> None:
        # raise a TypeError if self.original_filename is not a valid path
        # we also want to expand '$USER' and '~' because os.path.exists('~') is always False
        filename: str = os.path.expanduser(self.original_filename)

        if not os.path.exists(filename):
            raise FileNotFoundError(rf"No such file or directory '{filename!s}'")
        if self.root_folder is None:
            raise ValueError(
                f"Could not determine output directory from '{filename!s}'\n"
                f"Expected a directory name of form {OUTPUT_DIR_EXP!r} "
                "containing an info_*.txt file and amr_* files."
            )

        # This last case is (erroneously ?) marked as unreachable by mypy
        # If/when this bug is fixed upstream, mypy will warn that the unused
        # 'type: ignore' comment can be removed
        if self.info_fname is None:  # type: ignore [unreachable]
            raise ValueError(f"Failed to detect info file from '{filename!s}'")

    @property
    def is_valid(self) -> bool:
        try:
            self.validate()
        except (TypeError, FileNotFoundError, ValueError):
            return False
        else:
            return True

    @staticmethod
    def check_standard_files(folder, iout):
        """Return True if the folder contains an amr file and the info file."""
        # Check that the "amr_" and "info_" files exist
        ok = (folder / f"amr_{iout}.out00001").is_file()
        ok &= (folder / f"info_{iout}.txt").is_file()
        return ok

    @staticmethod
    def _match_output_and_group(
        path: Path,
    ) -> Tuple[Path, Optional[Path], Optional[str]]:

        # Make sure we work with a directory of the form `output_XXXXX`
        for p in (path, path.parent):
            match = OUTPUT_DIR_RE.match(p.name)

            if match:
                path = p
                break

        if match is None:
            return path, None, None

        iout = match.group(1)

        # See whether a folder named `group_YYYYY` exists
        group_dir = path / "group_00001"
        if group_dir.is_dir():
            return path, group_dir, iout
        else:
            return path, None, iout

    @classmethod
    def test_with_folder_name(
        cls, output_dir: Path
    ) -> Tuple[bool, Optional[Path], Optional[Path], Optional[Path]]:
        output_dir, group_dir, iout = cls._match_output_and_group(output_dir)
        ok = output_dir.is_dir() and iout is not None

        info_fname: Optional[Path]

        if ok:
            parent_dir = group_dir or output_dir
            ok &= cls.check_standard_files(parent_dir, iout)
            info_fname = parent_dir / f"info_{iout}.txt"
        else:
            info_fname = None

        return ok, output_dir, group_dir, info_fname

    @classmethod
    def test_with_standard_file(
        cls, filename: Path
    ) -> Tuple[bool, Optional[Path], Optional[Path], Optional[Path]]:
        output_dir, group_dir, iout = cls._match_output_and_group(filename.parent)
        ok = (
            filename.is_file()
            and STANDARD_FILE_RE.match(filename.name) is not None
            and iout is not None
        )

        info_fname: Optional[Path]

        if ok:
            parent_dir = group_dir or output_dir
            ok &= cls.check_standard_files(parent_dir, iout)
            info_fname = parent_dir / f"info_{iout}.txt"
        else:
            info_fname = None

        return ok, output_dir, group_dir, info_fname


class RAMSESDomainFile:
    _last_mask = None
    _last_selector_id = None

    def __init__(self, ds, domain_id):
        self.ds = ds
        self.domain_id = domain_id

        num = os.path.basename(ds.parameter_filename).split(".")[0].split("_")[1]
        rootdir = ds.root_folder
        basedir = os.path.abspath(os.path.dirname(ds.parameter_filename))
        basename = "%s/%%s_%s.out%05i" % (basedir, num, domain_id)
        part_file_descriptor = f"{basedir}/part_file_descriptor.txt"
        if ds.num_groups > 0:
            igroup = ((domain_id - 1) // ds.group_size) + 1
            basename = "%s/group_%05i/%%s_%s.out%05i" % (
                rootdir,
                igroup,
                num,
                domain_id,
            )
        else:
            basename = "%s/%%s_%s.out%05i" % (basedir, num, domain_id)
        for t in ["grav", "amr"]:
            setattr(self, f"{t}_fn", basename % t)
        self._part_file_descriptor = part_file_descriptor
        self._read_amr_header()

        # Autodetect field files
        field_handlers = [FH(self) for FH in get_field_handlers() if FH.any_exist(ds)]
        self.field_handlers = field_handlers
        for fh in field_handlers:
            mylog.debug("Detected fluid type %s in domain_id=%s", fh.ftype, domain_id)
            fh.detect_fields(ds)
            # self._add_ftype(fh.ftype)

        # Autodetect particle files
        particle_handlers = [
            PH(self) for PH in get_particle_handlers() if PH.any_exist(ds)
        ]
        self.particle_handlers = particle_handlers
        for ph in particle_handlers:
            mylog.debug(
                "Detected particle type %s in domain_id=%s", ph.ptype, domain_id
            )
            ph.read_header()
            # self._add_ptype(ph.ptype)

        # Load the AMR structure
        self._read_amr()

    _hydro_offset = None
    _level_count = None

    def __repr__(self):
        return "RAMSESDomainFile: %i" % self.domain_id

    @property
    def level_count(self):
        lvl_count = None
        for fh in self.field_handlers:
            fh.offset
            if lvl_count is None:
                lvl_count = fh.level_count.copy()
            else:
                lvl_count += fh._level_count
        return lvl_count

    @property
    def amr_file(self):
        if hasattr(self, "_amr_file"):
            self._amr_file.seek(0)
            return self._amr_file

        f = fpu(self.amr_fn)
        self._amr_file = f
        f.seek(0)
        return f

    def _read_amr_header(self):
        hvals = {}
        f = self.amr_file
        f.seek(0)

        for header in ramses_header(hvals):
            hvals.update(f.read_attrs(header))
        # For speedup, skip reading of 'headl' and 'taill'
        f.skip(2)
        hvals["numbl"] = f.read_vector("i")

        # That's the header, now we skip a few.
        hvals["numbl"] = np.array(hvals["numbl"]).reshape(
            (hvals["nlevelmax"], hvals["ncpu"])
        )
        f.skip()
        if hvals["nboundary"] > 0:
            f.skip(2)
            self.ngridbound = f.read_vector("i").astype("int64")
        else:
            self.ngridbound = np.zeros(hvals["nlevelmax"], dtype="int64")
        free_mem = f.read_attrs((("free_mem", 5, "i"),))  # NOQA
        ordering = f.read_vector("c")  # NOQA
        f.skip(4)
        # Now we're at the tree itself
        # Now we iterate over each level and each CPU.
        self.amr_header = hvals
        # update levelmax
        force_max_level, convention = self.ds._force_max_level
        if convention == "yt":
            force_max_level += self.ds.min_level + 1
        self.amr_header["nlevelmax"] = min(
            force_max_level, self.amr_header["nlevelmax"]
        )
        self.amr_offset = f.tell()
        self.local_oct_count = hvals["numbl"][
            self.ds.min_level :, self.domain_id - 1
        ].sum()
        self.total_oct_count = hvals["numbl"][self.ds.min_level :, :].sum(axis=0)

    def _read_amr(self):
        """Open the oct file, read in octs level-by-level.
        For each oct, only the position, index, level and domain
        are needed - its position in the octree is found automatically.
        The most important is finding all the information to feed
        oct_handler.add
        """
        self.oct_handler = RAMSESOctreeContainer(
            self.ds.domain_dimensions / 2,
            self.ds.domain_left_edge,
            self.ds.domain_right_edge,
        )
        root_nodes = self.amr_header["numbl"][self.ds.min_level, :].sum()
        self.oct_handler.allocate_domains(self.total_oct_count, root_nodes)
        mylog.debug(
            "Reading domain AMR % 4i (%0.3e, %0.3e)",
            self.domain_id,
            self.total_oct_count.sum(),
            self.ngridbound.sum(),
        )

        f = self.amr_file
        f.seek(self.amr_offset)

        min_level = self.ds.min_level
        max_level = read_amr(
            f, self.amr_header, self.ngridbound, min_level, self.oct_handler
        )

        self.max_level = max_level
        self.oct_handler.finalize()

        # Close AMR file
        f.close()

    def included(self, selector):
        if getattr(selector, "domain_id", None) is not None:
            return selector.domain_id == self.domain_id
        domain_ids = self.oct_handler.domain_identify(selector)
        return self.domain_id in domain_ids


class RAMSESDomainSubset(OctreeSubset):

    _domain_offset = 1
    _block_order = "F"

    _base_domain = None

    def __init__(
        self,
        base_region,
        domain,
        ds,
        num_zones=2,
        num_ghost_zones=0,
        base_grid=None,
    ):
        super().__init__(base_region, domain, ds, num_zones, num_ghost_zones)

        self._base_grid = base_grid

        if num_ghost_zones > 0:
            if not all(ds.periodicity):
                mylog.warning(
                    "Ghost zones will wrongly assume the domain to be periodic."
                )
            # Create a base domain *with no self._base_domain.fwidth
            base_domain = RAMSESDomainSubset(ds.all_data(), domain, ds, num_zones)
            self._base_domain = base_domain
        elif num_ghost_zones < 0:
            raise RuntimeError(
                "Cannot initialize a domain subset with a negative number "
                "of ghost zones, was called with num_ghost_zones=%s" % num_ghost_zones
            )

    def _fill_no_ghostzones(self, fd, fields, selector, file_handler):
        ndim = self.ds.dimensionality
        # Here we get a copy of the file, which we skip through and read the
        # bits we want.
        oct_handler = self.oct_handler
        all_fields = [f for ft, f in file_handler.field_list]
        fields = [f for ft, f in fields]
        data = {}
        cell_count = selector.count_oct_cells(self.oct_handler, self.domain_id)

        levels, cell_inds, file_inds = self.oct_handler.file_index_octs(
            selector, self.domain_id, cell_count
        )

        # Initializing data container
        for field in fields:
            data[field] = np.zeros(cell_count, "float64")

        cpu_list = [self.domain_id - 1]
        fill_hydro(
            fd,
            file_handler.offset,
            file_handler.level_count,
            cpu_list,
            levels,
            cell_inds,
            file_inds,
            ndim,
            all_fields,
            fields,
            data,
            oct_handler,
        )
        return data

    def _fill_with_ghostzones(
        self, fd, fields, selector, file_handler, num_ghost_zones
    ):
        ndim = self.ds.dimensionality
        ncpu = self.ds.parameters["ncpu"]
        # Here we get a copy of the file, which we skip through and read the
        # bits we want.
        oct_handler = self.oct_handler
        all_fields = [f for ft, f in file_handler.field_list]
        fields = [f for ft, f in fields]
        tr = {}

        cell_count = (
            selector.count_octs(self.oct_handler, self.domain_id) * self.nz**ndim
        )

        gz_cache = getattr(self, "_ghost_zone_cache", None)
        if gz_cache:
            levels, cell_inds, file_inds, domains = gz_cache
        else:
            gz_cache = (
                levels,
                cell_inds,
                file_inds,
                domains,
            ) = self.oct_handler.file_index_octs_with_ghost_zones(
                selector, self.domain_id, cell_count, self._num_ghost_zones
            )
            self._ghost_zone_cache = gz_cache

        # Initializing data container
        for field in fields:
            tr[field] = np.zeros(cell_count, "float64")
        cpu_list = list(range(ncpu))
        fill_hydro(
            fd,
            file_handler.offset,
            file_handler.level_count,
            cpu_list,
            levels,
            cell_inds,
            file_inds,
            ndim,
            all_fields,
            fields,
            tr,
            oct_handler,
            domains=domains,
        )
        return tr

    @property
    def fwidth(self):
        fwidth = super().fwidth
        if self._num_ghost_zones > 0:
            fwidth = fwidth.reshape(-1, 8, 3)
            n_oct = fwidth.shape[0]
            # new_fwidth contains the fwidth of the oct+ghost zones
            # this is a constant array in each oct, so we simply copy
            # the oct value using numpy fancy-indexing
            new_fwidth = np.zeros((n_oct, self.nz**3, 3), dtype=fwidth.dtype)
            new_fwidth[:, :, :] = fwidth[:, 0:1, :]
            fwidth = new_fwidth.reshape(-1, 3)
        return fwidth

    @property
    def fcoords(self):
        num_ghost_zones = self._num_ghost_zones
        if num_ghost_zones == 0:
            return super().fcoords

        oh = self.oct_handler

        indices = oh.fill_index(self.selector).reshape(-1, 8)
        oct_inds, cell_inds = oh.fill_octcellindex_neighbours(
            self.selector, self._num_ghost_zones
        )

        N_per_oct = self.nz**3
        oct_inds = oct_inds.reshape(-1, N_per_oct)
        cell_inds = cell_inds.reshape(-1, N_per_oct)

        inds = indices[oct_inds, cell_inds]

        fcoords = self.ds.arr(oh.fcoords(self.selector)[inds].reshape(-1, 3), "unitary")

        return fcoords

    def fill(self, fd, fields, selector, file_handler):
        if self._num_ghost_zones == 0:
            return self._fill_no_ghostzones(fd, fields, selector, file_handler)
        else:
            return self._fill_with_ghostzones(
                fd, fields, selector, file_handler, self._num_ghost_zones
            )

    def retrieve_ghost_zones(self, ngz, fields, smoothed=False):
        if smoothed:
            mylog.warning(
                "%s.retrieve_ghost_zones was called with the "
                "`smoothed` argument set to True. This is not supported, "
                "ignoring it.",
                self,
            )
            smoothed = False

        _subset_with_gz = getattr(self, "_subset_with_gz", {})

        try:
            new_subset = _subset_with_gz[ngz]
            mylog.debug(
                "Reusing previous subset with %s ghost zones for domain %s",
                ngz,
                self.domain_id,
            )
        except KeyError:
            new_subset = RAMSESDomainSubset(
                self.base_region,
                self.domain,
                self.ds,
                num_ghost_zones=ngz,
                base_grid=self,
            )
            _subset_with_gz[ngz] = new_subset

        # Cache the fields
        new_subset.get_data(fields)
        self._subset_with_gz = _subset_with_gz

        return new_subset


class RAMSESIndex(OctreeIndex):
    def __init__(self, ds, dataset_type="ramses"):
        self.fluid_field_list = ds._fields_in_file
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self.max_level = None

        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    def _initialize_oct_handler(self):
        if self.ds._bbox is not None:
            cpu_list = get_cpu_list(self.dataset, self.dataset._bbox)
        else:
            cpu_list = range(self.dataset["ncpu"])

        self.domains = [RAMSESDomainFile(self.dataset, i + 1) for i in cpu_list]
        total_octs = sum(
            dom.local_oct_count for dom in self.domains  # + dom.ngridbound.sum()
        )
        force_max_level, convention = self.ds._force_max_level
        if convention == "yt":
            force_max_level += self.ds.min_level + 1
        self.max_level = min(
            force_max_level, max(dom.max_level for dom in self.domains)
        )
        self.num_grids = total_octs

    def _detect_output_fields(self):
        dsl = set()

        # Get the detected particle fields
        for domain in self.domains:
            for ph in domain.particle_handlers:
                dsl.update(set(ph.field_offsets.keys()))

        self.particle_field_list = list(dsl)
        cosmo = self.ds.cosmological_simulation
        for f in self.particle_field_list[:]:
            if f[1] == "particle_birth_time" and cosmo:
                self.particle_field_list.append((f[0], "conformal_birth_time"))

        # Get the detected fields
        dsl = set()
        for fh in self.domains[0].field_handlers:
            dsl.update(set(fh.field_list))
        self.fluid_field_list = list(dsl)

        self.field_list = self.particle_field_list + self.fluid_field_list

    def _identify_base_chunk(self, dobj):
        if getattr(dobj, "_chunk_info", None) is None:
            domains = [dom for dom in self.domains if dom.included(dobj.selector)]
            base_region = getattr(dobj, "base_region", dobj)
            if len(domains) > 1:
                mylog.debug("Identified %s intersecting domains", len(domains))
            subsets = [
                RAMSESDomainSubset(
                    base_region,
                    domain,
                    self.dataset,
                    num_ghost_zones=dobj._num_ghost_zones,
                )
                for domain in domains
            ]
            dobj._chunk_info = subsets
        dobj._current_chunk = list(self._chunk_all(dobj))[0]

    def _chunk_all(self, dobj):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        yield YTDataChunk(dobj, "all", oobjs, None)

    def _chunk_spatial(self, dobj, ngz, sort=None, preload_fields=None):
        sobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for og in sobjs:
            if ngz > 0:
                g = og.retrieve_ghost_zones(ngz, [], smoothed=True)
            else:
                g = og
            yield YTDataChunk(dobj, "spatial", [g], None)

    def _chunk_io(self, dobj, cache=True, local_only=False):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for subset in oobjs:
            yield YTDataChunk(dobj, "io", [subset], None, cache=cache)

    def _initialize_level_stats(self):
        levels = sum(dom.level_count for dom in self.domains)
        desc = {"names": ["numcells", "level"], "formats": ["int64"] * 2}
        max_level = self.dataset.min_level + self.dataset.max_level + 2
        self.level_stats = blankRecordArray(desc, max_level)
        self.level_stats["level"] = [i for i in range(max_level)]
        self.level_stats["numcells"] = [0 for i in range(max_level)]
        for level in range(self.dataset.min_level + 1):
            self.level_stats[level + 1]["numcells"] = 2 ** (
                level * self.dataset.dimensionality
            )
        for level in range(self.max_level + 1):
            self.level_stats[level + self.dataset.min_level + 1]["numcells"] = levels[
                :, level
            ].sum()

    def _get_particle_type_counts(self):
        npart = 0
        npart = {k: 0 for k in self.ds.particle_types if k != "all"}
        for dom in self.domains:
            for fh in dom.particle_handlers:
                count = fh.local_particle_count
                npart[fh.ptype] += count

        return npart

    def print_stats(self):
        """
        Prints out (stdout) relevant information about the simulation

        This function prints information based on the fluid on the grids,
        and therefore does not work for DM only runs.
        """
        if not self.fluid_field_list:
            print("This function is not implemented for DM only runs")
            return

        self._initialize_level_stats()

        header = "{:>3}\t{:>14}\t{:>14}".format("level", "# cells", "# cells^3")
        print(header)
        print(f"{len(header.expandtabs()) * '-'}")
        for level in range(self.dataset.min_level + self.dataset.max_level + 2):
            print(
                "% 3i\t% 14i\t% 14i"
                % (
                    level,
                    self.level_stats["numcells"][level],
                    np.ceil(self.level_stats["numcells"][level] ** (1.0 / 3)),
                )
            )
        print("-" * 46)
        print("   \t% 14i" % (self.level_stats["numcells"].sum()))
        print("\n")

        dx = self.get_smallest_dx()
        try:
            print(f"z = {self.dataset.current_redshift:0.8f}")
        except Exception:
            pass
        print(
            "t = %0.8e = %0.8e s = %0.8e years"
            % (
                self.ds.current_time.in_units("code_time"),
                self.ds.current_time.in_units("s"),
                self.ds.current_time.in_units("yr"),
            )
        )
        print("\nSmallest Cell:")
        for item in ("Mpc", "pc", "AU", "cm"):
            print(f"\tWidth: {dx.in_units(item):0.3e} {item}")


class RAMSESDataset(Dataset):
    _index_class = RAMSESIndex
    _field_info_class = RAMSESFieldInfo
    gamma = 1.4  # This will get replaced on hydro_fn open

    def __init__(
        self,
        filename,
        dataset_type="ramses",
        fields=None,
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
        extra_particle_fields=None,
        cosmological=None,
        bbox=None,
        max_level=None,
        max_level_convention=None,
        default_species_fields=None,
    ):
        # Here we want to initiate a traceback, if the reader is not built.
        if isinstance(fields, str):
            fields = field_aliases[fields]
        """
        fields:
        An array of hydro variable fields in order of position in the
        hydro_XXXXX.outYYYYY file. If set to None, will try a default set of fields.

        extra_particle_fields:
        An array of extra particle variables in order of position in the
        particle_XXXXX.outYYYYY file.

        cosmological:
        If set to None, automatically detect cosmological simulation.
        If a boolean, force its value.
        """

        self._fields_in_file = fields
        # By default, extra fields have not triggered a warning
        self._warned_extra_fields = defaultdict(lambda: False)
        self._extra_particle_fields = extra_particle_fields
        self.force_cosmological = cosmological
        self._bbox = bbox

        self._force_max_level = self._sanitize_max_level(
            max_level, max_level_convention
        )

        file_handler = RAMSESFileSanitizer(filename)

        # ensure validation happens even if the class is instanciated
        # directly rather than from yt.load
        file_handler.validate()

        # Sanitize the filename
        info_fname = file_handler.info_fname

        if file_handler.group_name is not None:
            self.num_groups = len(
                [_ for _ in file_handler.root_folder.glob("group_?????") if _.is_dir()]
            )
        else:
            self.num_groups = 0
        self.root_folder = file_handler.root_folder

        Dataset.__init__(
            self,
            info_fname,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

        # Add the particle types
        ptypes = []
        for PH in get_particle_handlers():
            if PH.any_exist(self):
                ptypes.append(PH.ptype)

        ptypes = tuple(ptypes)
        self.particle_types = self.particle_types_raw = ptypes

        # Add the fluid types
        for FH in get_field_handlers():
            FH.purge_detected_fields(self)
            if FH.any_exist(self):
                self.fluid_types += (FH.ftype,)

        self.storage_filename = storage_filename

    @staticmethod
    def _sanitize_max_level(max_level, max_level_convention):
        # NOTE: the initialisation of the dataset class sets
        #       self.min_level _and_ requires force_max_level
        #       to be set, so we cannot convert from to yt/ramses
        #       conventions
        if max_level is None and max_level_convention is None:
            return (2**999, "yt")

        # Check max_level is a valid, positive integer
        if not isinstance(max_level, (int, np.integer)):
            raise TypeError(
                f"Expected `max_level` to be a positive integer, got {max_level} "
                f"with type {type(max_level)} instead."
            )
        if max_level < 0:
            raise ValueError(
                f"Expected `max_level` to be a positive integer, got {max_level} "
                "instead."
            )

        # Check max_level_convention is set and acceptable
        if max_level_convention is None:
            raise ValueError(
                f"Received `max_level`={max_level}, but no `max_level_convention`. "
                "Valid conventions are 'yt' and 'ramses'."
            )
        if max_level_convention not in ("ramses", "yt"):
            raise ValueError(
                f"Invalid convention {max_level_convention}. "
                "Valid choices are 'yt' and 'ramses'."
            )
        return (max_level, max_level_convention)

    def create_field_info(self, *args, **kwa):
        """Extend create_field_info to add the particles types."""
        super().create_field_info(*args, **kwa)
        # Register particle filters
        if ("io", "particle_family") in self.field_list:
            for fname, value in particle_families.items():

                def loc(val):
                    def closure(pfilter, data):
                        filter = data[(pfilter.filtered_type, "particle_family")] == val
                        return filter

                    return closure

                add_particle_filter(
                    fname, loc(value), filtered_type="io", requires=["particle_family"]
                )

            for k in particle_families.keys():
                mylog.info("Adding particle_type: %s", k)
                self.add_particle_filter(f"{k}")

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        # loading the units from the info file
        boxlen = self.parameters["boxlen"]
        length_unit = self.parameters["unit_l"]
        density_unit = self.parameters["unit_d"]
        time_unit = self.parameters["unit_t"]

        # calculating derived units (except velocity and temperature, done below)
        mass_unit = density_unit * length_unit**3
        magnetic_unit = np.sqrt(4 * np.pi * mass_unit / (time_unit**2 * length_unit))
        pressure_unit = density_unit * (length_unit / time_unit) ** 2

        # TODO:
        # Generalize the temperature field to account for ionization
        # For now assume an atomic ideal gas with cosmic abundances (x_H = 0.76)
        mean_molecular_weight_factor = _X**-1

        setdefaultattr(self, "density_unit", self.quan(density_unit, "g/cm**3"))
        setdefaultattr(self, "magnetic_unit", self.quan(magnetic_unit, "gauss"))
        setdefaultattr(self, "pressure_unit", self.quan(pressure_unit, "dyne/cm**2"))
        setdefaultattr(self, "time_unit", self.quan(time_unit, "s"))
        setdefaultattr(self, "mass_unit", self.quan(mass_unit, "g"))
        setdefaultattr(
            self, "velocity_unit", self.quan(length_unit, "cm") / self.time_unit
        )
        temperature_unit = (
            self.velocity_unit**2 * mp * mean_molecular_weight_factor / kb
        )
        setdefaultattr(self, "temperature_unit", temperature_unit.in_units("K"))

        # Only the length unit get scales by a factor of boxlen
        setdefaultattr(self, "length_unit", self.quan(length_unit * boxlen, "cm"))

    def _parse_parameter_file(self):
        # hardcoded for now
        # These should be explicitly obtained from the file, but for now that
        # will wait until a reorganization of the source tree and better
        # generalization.
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "ramses"
        self.parameters["Time"] = 1.0  # default unit is 1...

        # We now execute the same logic Oliver's code does
        rheader = {}

        def read_rhs(f, cast):
            line = f.readline().strip()

            if line and "=" in line:
                key, val = (_.strip() for _ in line.split("="))
                rheader[key] = cast(val)
                return key
            else:
                return None

        def cast_a_else_b(cast_a, cast_b):
            def caster(val):
                try:
                    return cast_a(val)
                except ValueError:
                    return cast_b(val)

            return caster

        with open(self.parameter_filename) as f:
            # Standard: first six are ncpu, ndim, levelmin, levelmax, ngridmax, nstep_coarse
            for _ in range(6):
                read_rhs(f, int)
            f.readline()
            # Standard: next 11 are boxlen, time, aexp, h0, omega_m, omega_l, omega_k, omega_b, unit_l, unit_d, unit_t
            for _ in range(11):
                key = read_rhs(f, float)

            # Read non standard extra fields until hitting the ordering type
            while key != "ordering type":
                key = read_rhs(f, cast_a_else_b(float, str))

            # This next line deserves some comment.  We specify a min_level that
            # corresponds to the minimum level in the RAMSES simulation.  RAMSES is
            # one-indexed, but it also does refer to the *oct* dimensions -- so
            # this means that a levelmin of 1 would have *1* oct in it.  So a
            # levelmin of 2 would have 8 octs at the root mesh level.
            self.min_level = rheader["levelmin"] - 1
            # Now we read the hilbert indices
            self.hilbert_indices = {}
            if rheader["ordering type"] == "hilbert":
                f.readline()  # header
                for _ in range(rheader["ncpu"]):
                    dom, mi, ma = f.readline().split()
                    self.hilbert_indices[int(dom)] = (float(mi), float(ma))

        if rheader["ordering type"] != "hilbert" and self._bbox is not None:
            raise NotImplementedError(
                "The ordering %s is not compatible with the `bbox` argument."
                % rheader["ordering type"]
            )
        self.parameters.update(rheader)
        self.domain_left_edge = np.zeros(3, dtype="float64")
        self.domain_dimensions = np.ones(3, dtype="int32") * 2 ** (self.min_level + 1)
        self.domain_right_edge = np.ones(3, dtype="float64")
        # This is likely not true, but it's not clear
        # how to determine the boundary conditions
        self._periodicity = (True, True, True)

        if self.force_cosmological is not None:
            is_cosmological = self.force_cosmological
        else:
            # These conditions seem to always be true for non-cosmological datasets
            is_cosmological = not (
                rheader["time"] >= 0 and rheader["H0"] == 1 and rheader["aexp"] == 1
            )

        if not is_cosmological:
            self.cosmological_simulation = 0
            self.current_redshift = 0
            self.hubble_constant = 0
            self.omega_matter = 0
            self.omega_lambda = 0
        else:
            self.cosmological_simulation = 1
            self.current_redshift = (1.0 / rheader["aexp"]) - 1.0
            self.omega_lambda = rheader["omega_l"]
            self.omega_matter = rheader["omega_m"]
            self.hubble_constant = rheader["H0"] / 100.0  # This is H100

        force_max_level, convention = self._force_max_level
        if convention == "yt":
            force_max_level += self.min_level + 1
        self.max_level = min(force_max_level, rheader["levelmax"]) - self.min_level - 1

        if self.cosmological_simulation == 0:
            self.current_time = self.parameters["time"]
        else:
            self.tau_frw, self.t_frw, self.dtau, self.n_frw, self.time_tot = friedman(
                self.omega_matter,
                self.omega_lambda,
                1.0 - self.omega_matter - self.omega_lambda,
            )

            age = self.parameters["time"]
            iage = 1 + int(10.0 * age / self.dtau)
            iage = np.min([iage, self.n_frw // 2 + (iage - self.n_frw // 2) // 10])

            try:
                self.time_simu = self.t_frw[iage] * (age - self.tau_frw[iage - 1]) / (
                    self.tau_frw[iage] - self.tau_frw[iage - 1]
                ) + self.t_frw[iage - 1] * (age - self.tau_frw[iage]) / (
                    self.tau_frw[iage - 1] - self.tau_frw[iage]
                )

                self.current_time = (
                    (self.time_tot + self.time_simu)
                    / (self.hubble_constant * 1e7 / cm_per_mpc)
                    / self.parameters["unit_t"]
                )
            except IndexError:
                mylog.warning(
                    "Yt could not convert conformal time to physical time. "
                    "Yt will assume the simulation is *not* cosmological."
                )
                self.cosmological_simulation = 0
                self.current_time = self.parameters["time"]

        if self.num_groups > 0:
            self.group_size = rheader["ncpu"] // self.num_groups

        # Read namelist.txt file (if any)
        self.read_namelist()

    def read_namelist(self):
        """Read the namelist.txt file in the output folder, if present"""
        namelist_file = os.path.join(self.root_folder, "namelist.txt")
        if os.path.exists(namelist_file):
            try:
                with open(namelist_file) as f:
                    nml = f90nml.read(f)
            except ImportError as e:
                nml = f"An error occurred when reading the namelist: {str(e)}"
            except (ValueError, StopIteration, AssertionError) as err:
                # Note: f90nml may raise a StopIteration, a ValueError or an AssertionError if
                # the namelist is not valid.
                mylog.warning(
                    "Could not parse `namelist.txt` file as it was malformed:",
                    exc_info=err,
                )
                return

            self.parameters["namelist"] = nml

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return RAMSESFileSanitizer(filename).is_valid



import os

from yt.utilities.lib.cykdtree.kdtree import PyKDTree
from yt.utilities.lib.cykdtree.plot import plot2D_serial
from yt.utilities.lib.cykdtree.tests import make_points


def test_plot2D_serial():
    fname_test = "test_plot2D_serial.png"
    pts, le, re, ls = make_points(100, 2)
    tree = PyKDTree(pts, le, re, leafsize=ls)
    axs = plot2D_serial(
        tree, pts, title="Serial Test", plotfile=fname_test, label_boxes=True
    )
    os.remove(fname_test)
    # plot2D_serial(tree, pts, axs=axs)
    del axs



import numpy as np

from yt.utilities.logger import ytLogger as mylog


def parse_unit_dimension(unit_dimension):
    r"""Transforms an openPMD unitDimension into a string.

    Parameters
    ----------
    unit_dimension : array_like
        integer array of length 7 with one entry for the dimensional component of every
        SI unit

        [0] length L,
        [1] mass M,
        [2] time T,
        [3] electric current I,
        [4] thermodynamic temperature theta,
        [5] amount of substance N,
        [6] luminous intensity J

    References
    ----------

    https://github.com/openPMD/openPMD-standard/blob/latest/STANDARD.md#unit-systems-and-dimensionality


    Returns
    -------
    str

    Examples
    --------
    >>> velocity = [1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0]
    >>> print(parse_unit_dimension(velocity))
    'm**1*s**-1'

    >>> magnetic_field = [0.0, 1.0, -2.0, -1.0, 0.0, 0.0, 0.0]
    >>> print(parse_unit_dimension(magnetic_field))
    'kg**1*s**-2*A**-1'
    """
    if len(unit_dimension) != 7:
        mylog.error("SI must have 7 base dimensions!")
    unit_dimension = np.asarray(unit_dimension, dtype="int64")
    dim = []
    si = ["m", "kg", "s", "A", "C", "mol", "cd"]
    for i in np.arange(7):
        if unit_dimension[i] != 0:
            dim.append(f"{si[i]}**{unit_dimension[i]}")
    return "*".join(dim)


def is_const_component(record_component):
    """Determines whether a group or dataset in the HDF5 file is constant.

    Parameters
    ----------
    record_component : h5py.Group or h5py.Dataset

    Returns
    -------
    bool
        True if constant, False otherwise

    References
    ----------
    .. https://github.com/openPMD/openPMD-standard/blob/latest/STANDARD.md,
       section 'Constant Record Components'
    """
    return "value" in record_component.attrs.keys()


def get_component(group, component_name, index=0, offset=None):
    """Grabs a dataset component from a group as a whole or sliced.

    Parameters
    ----------
    group : h5py.Group
    component_name : str
        relative path of the component in the group
    index : int, optional
        first entry along the first axis to read
    offset : int, optional
        number of entries to read
        if not supplied, every entry after index is returned

    Notes
    -----
    This scales every entry of the component with the respective "unitSI".

    Returns
    -------
    ndarray
        (N,) 1D in case of particle data
        (O,P,Q) 1D/2D/3D in case of mesh data
    """
    record_component = group[component_name]
    unit_si = record_component.attrs["unitSI"]
    if is_const_component(record_component):
        shape = np.asarray(record_component.attrs["shape"])
        if offset is None:
            shape[0] -= index
        else:
            shape[0] = offset
        # component is constant, craft an array by hand
        return np.full(shape, record_component.attrs["value"] * unit_si)
    else:
        if offset is not None:
            offset += index
        # component is a dataset, return it (possibly masked)
        return np.multiply(record_component[index:offset], unit_si)



cdef class HaloParticlesSelector(SelectorObject):
    cdef public object base_source
    cdef SelectorObject base_selector
    cdef object pind
    cdef public np.int64_t halo_id
    def __init__(self, dobj):
        self.base_source = dobj.base_source
        self.base_selector = self.base_source.selector
        self.pind = dobj.particle_indices

    def _hash_vals(self):
        return ("halo_particles", self.halo_id)

halo_particles_selector = HaloParticlesSelector



"""
Some simple operations for operating on ragged arrays



"""


import numpy as np

cimport cython
cimport numpy as np


cdef fused numpy_dt:
    np.float32_t
    np.float64_t
    np.int32_t
    np.int64_t

cdef numpy_dt r_min(numpy_dt a, numpy_dt b):
    if a < b: return a
    return b

cdef numpy_dt r_max(numpy_dt a, numpy_dt b):
    if a > b: return a
    return b

cdef numpy_dt r_add(numpy_dt a, numpy_dt b):
    return a + b

cdef numpy_dt r_subtract(numpy_dt a, numpy_dt b):
    return a - b

cdef numpy_dt r_multiply(numpy_dt a, numpy_dt b):
    return a * b

@cython.cdivision(True)
cdef numpy_dt r_divide(numpy_dt a, numpy_dt b):
    return a / b

def index_unop(np.ndarray[numpy_dt, ndim=1] values,
              np.ndarray[np.int64_t, ndim=1] indices,
              np.ndarray[np.int64_t, ndim=1] sizes,
              operation):
    cdef numpy_dt mi, ma
    if numpy_dt == np.float32_t:
        dt = "float32"
        mi = np.finfo(dt).min
        ma = np.finfo(dt).max
    elif numpy_dt == np.float64_t:
        dt = "float64"
        mi = np.finfo(dt).min
        ma = np.finfo(dt).max
    elif numpy_dt == np.int32_t:
        dt = "int32"
        mi = np.iinfo(dt).min
        ma = np.iinfo(dt).max
    elif numpy_dt == np.int64_t:
        dt = "int64"
        mi = np.iinfo(dt).min
        ma = np.iinfo(dt).max
    cdef np.ndarray[numpy_dt] out_values = np.zeros(sizes.size, dtype=dt)
    cdef numpy_dt (*func)(numpy_dt a, numpy_dt b)
    # Now we figure out our function.  At present, we only allow addition and
    # multiplication, because they are commutative and easy to bootstrap.
    cdef numpy_dt ival, val
    if operation == "sum":
        ival = 0
        func = r_add
    elif operation == "prod":
        ival = 1
        func = r_multiply
    elif operation == "max":
        ival = mi
        func = r_max
    elif operation == "min":
        ival = ma
        func = r_min
    else:
        raise NotImplementedError
    cdef np.int64_t i, ind_ind, ind_arr
    ind_ind = 0
    for i in range(sizes.size):
        # Each entry in sizes is the size of the array
        val = ival
        for _ in range(sizes[i]):
            ind_arr = indices[ind_ind]
            val = func(val, values[ind_arr])
            ind_ind += 1
        out_values[i] = val
    return out_values



import numpy as np

from yt.funcs import mylog
from yt.utilities.io_handler import BaseParticleIOHandler
from yt.utilities.on_demand_imports import _requests as requests


class IOHandlerHTTPStream(BaseParticleIOHandler):
    _dataset_type = "http_particle_stream"
    _vector_fields = {"Coordinates": 3, "Velocity": 3, "Velocities": 3}

    def __init__(self, ds):
        self._url = ds.base_url
        # This should eventually manage the IO and cache it
        self.total_bytes = 0
        super().__init__(ds)

    def _open_stream(self, data_file, field):
        # This does not actually stream yet!
        ftype, fname = field
        s = f"{self._url}/{data_file.file_id}/{ftype}/{fname}"
        mylog.info("Loading URL %s", s)
        resp = requests.get(s)
        if resp.status_code != 200:
            raise RuntimeError
        self.total_bytes += len(resp.content)
        return resp.content

    def _identify_fields(self, data_file):
        f = []
        for ftype, fname in self.ds.parameters["field_list"]:
            f.append((str(ftype), str(fname)))
        return f, {}

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            for ptype in ptf:
                s = self._open_stream(data_file, (ptype, "Coordinates"))
                c = np.frombuffer(s, dtype="float64")
                c.shape = (c.shape[0] / 3.0, 3)
                yield ptype, (c[:, 0], c[:, 1], c[:, 2]), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            for ptype, field_list in sorted(ptf.items()):
                s = self._open_stream(data_file, (ptype, "Coordinates"))
                c = np.frombuffer(s, dtype="float64")
                c.shape = (c.shape[0] / 3.0, 3)
                mask = selector.select_points(c[:, 0], c[:, 1], c[:, 2], 0.0)
                del c
                if mask is None:
                    continue
                for field in field_list:
                    s = self._open_stream(data_file, (ptype, field))
                    c = np.frombuffer(s, dtype="float64")
                    if field in self._vector_fields:
                        c.shape = (c.shape[0] / 3.0, 3)
                    data = c[mask, ...]
                    yield (ptype, field), data

    def _count_particles(self, data_file):
        return self.ds.parameters["particle_count"][data_file.file_id]



import numpy as np

from yt import load_uniform_grid
from yt.testing import assert_almost_equal, assert_equal


def test_variable_dx():
    np.random.seed(0x4D3D3D3)

    data = {"density": np.random.random((128, 128, 128))}

    cell_widths = []
    for _ in range(3):
        cw = np.random.random(128)
        cw /= cw.sum()
        cell_widths.append(cw)

    ds = load_uniform_grid(
        data,
        [128, 128, 128],
        bbox=np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]]),
        cell_widths=cell_widths,
    )

    # We now check that we get all of our original cell widths back out, and
    # only those cell widths

    assert_equal(np.unique(ds.index.grids[0]["index", "dx"]).size, 128)
    assert_equal(ds.index.grids[0]["index", "dx"][:, 0, 0], cell_widths[0])

    assert_equal(np.unique(ds.index.grids[0]["index", "dx"]).size, 128)
    assert_equal(ds.index.grids[0]["index", "dy"][0, :, 0], cell_widths[1])

    assert_equal(np.unique(ds.index.grids[0]["index", "dx"]).size, 128)
    assert_equal(ds.index.grids[0]["index", "dz"][0, 0, :], cell_widths[2])

    assert_equal(np.unique(ds.index.grids[0]["index", "x"]).size, 128)
    center_x = np.add.accumulate(cell_widths[0]) - 0.5 * cell_widths[0]
    assert_equal(center_x, ds.index.grids[0]["index", "x"][:, 0, 0])

    assert_equal(np.unique(ds.index.grids[0]["index", "y"]).size, 128)
    center_y = np.add.accumulate(cell_widths[1]) - 0.5 * cell_widths[1]
    assert_equal(center_y, ds.index.grids[0]["index", "y"][0, :, 0])

    assert_equal(np.unique(ds.index.grids[0]["index", "z"]).size, 128)
    center_z = np.add.accumulate(cell_widths[2]) - 0.5 * cell_widths[2]
    assert_equal(center_z, ds.index.grids[0]["index", "z"][0, 0, :])

    assert_almost_equal(ds.r[:].sum(("index", "cell_volume")), ds.domain_width.prod())

    for ax in "xyz":
        dd = ds.all_data()
        p = dd.integrate("ones", axis=ax)
        assert_almost_equal(p["index", "ones"].max().d, 1.0)
        assert_almost_equal(p["index", "ones"].min().d, 1.0)



import numpy as np

from yt.frontends.gadget.api import IOHandlerGadgetHDF5
from yt.utilities.on_demand_imports import _h5py as h5py


class IOHandlerArepoHDF5(IOHandlerGadgetHDF5):
    _dataset_type = "arepo_hdf5"

    def _generate_smoothing_length(self, index):
        # This is handled below in _get_smoothing_length
        return

    def _get_smoothing_length(self, data_file, position_dtype, position_shape):
        ptype = self.ds._sph_ptypes[0]
        ind = int(ptype[-1])
        si, ei = data_file.start, data_file.end
        with h5py.File(data_file.filename, mode="r") as f:
            pcount = f["/Header"].attrs["NumPart_ThisFile"][ind].astype("int")
            pcount = np.clip(pcount - si, 0, ei - si)
            # Arepo cells do not have "smoothing lengths" by definition, so
            # we compute one here by finding the radius of the sphere
            # corresponding to the volume of the Voroni cell and multiplying
            # by a user-configurable smoothing factor.
            hsml = f[ptype]["Masses"][si:ei, ...] / f[ptype]["Density"][si:ei, ...]
            hsml *= 3.0 / (4.0 * np.pi)
            hsml **= 1.0 / 3.0
            hsml *= self.ds.smoothing_factor
            dt = hsml.dtype.newbyteorder("N")  # Native
            if position_dtype is not None and dt < position_dtype:
                dt = position_dtype
            return hsml.astype(dt)



import functools
import glob
import inspect
import os
import weakref
from functools import wraps
from typing import Optional, Type

import numpy as np
from more_itertools import always_iterable

from yt.config import ytcfg
from yt.data_objects.analyzer_objects import AnalysisTask, create_quantity_proxy
from yt.data_objects.particle_trajectories import ParticleTrajectories
from yt.funcs import is_sequence, mylog
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.exceptions import YTException
from yt.utilities.object_registries import (
    analysis_task_registry,
    data_object_registry,
    derived_quantity_registry,
    simulation_time_series_registry,
)
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    communication_system,
    parallel_objects,
    parallel_root_only,
)


class AnalysisTaskProxy:
    def __init__(self, time_series):
        self.time_series = time_series

    def __getitem__(self, key):
        task_cls = analysis_task_registry[key]

        @wraps(task_cls.__init__)
        def func(*args, **kwargs):
            task = task_cls(*args, **kwargs)
            return self.time_series.eval(task)

        return func

    def keys(self):
        return analysis_task_registry.keys()

    def __contains__(self, key):
        return key in analysis_task_registry


def get_ds_prop(propname):
    def _eval(params, ds):
        return getattr(ds, propname)

    cls = type(propname, (AnalysisTask,), dict(eval=_eval, _params=tuple()))
    return cls


attrs = (
    "refine_by",
    "dimensionality",
    "current_time",
    "domain_dimensions",
    "domain_left_edge",
    "domain_right_edge",
    "unique_identifier",
    "current_redshift",
    "cosmological_simulation",
    "omega_matter",
    "omega_lambda",
    "omega_radiation",
    "hubble_constant",
)


class TimeSeriesParametersContainer:
    def __init__(self, data_object):
        self.data_object = data_object

    def __getattr__(self, attr):
        if attr in attrs:
            return self.data_object.eval(get_ds_prop(attr)())
        raise AttributeError(attr)


class DatasetSeries:
    r"""The DatasetSeries object is a container of multiple datasets,
    allowing easy iteration and computation on them.

    DatasetSeries objects are designed to provide easy ways to access,
    analyze, parallelize and visualize multiple datasets sequentially.  This is
    primarily expressed through iteration, but can also be constructed via
    analysis tasks (see :ref:`time-series-analysis`).

    Note that contained datasets are lazily loaded and weakly referenced. This means
    that in order to perform follow-up operations on data it's best to define handles on
    these datasets during iteration.

    Parameters
    ----------
    outputs : list of filenames, or pattern
        A list of filenames, for instance ["DD0001/DD0001", "DD0002/DD0002"],
        or a glob pattern (i.e. containing wildcards '[]?!*') such as "DD*/DD*.index".
        In the latter case, results are sorted automatically.
        Filenames and patterns can be of type str, os.Pathlike or bytes.
    parallel : True, False or int
        This parameter governs the behavior when .piter() is called on the
        resultant DatasetSeries object.  If this is set to False, the time
        series will not iterate in parallel when .piter() is called.  If
        this is set to either True, one processor will be allocated for
        each iteration of the loop. If this is set to an integer, the loop
        will be parallelized over this many workgroups. It the integer
        value is less than the total number of available processors,
        more than one processor will be allocated to a given loop iteration,
        causing the functionality within the loop to be run in parallel.
    setup_function : callable, accepts a ds
        This function will be called whenever a dataset is loaded.
    mixed_dataset_types : True or False, default False
        Set to True if the DatasetSeries will load different dataset types, set
        to False if loading dataset of a single type as this will result in a
        considerable speed up from not having to figure out the dataset type.

    Examples
    --------

    >>> ts = DatasetSeries(
    ...     "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0[0-6][0-9]0"
    ... )
    >>> for ds in ts:
    ...     SlicePlot(ds, "x", ("gas", "density")).save()
    ...
    >>> def print_time(ds):
    ...     print(ds.current_time)
    ...
    >>> ts = DatasetSeries(
    ...     "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0[0-6][0-9]0",
    ...     setup_function=print_time,
    ... )
    ...
    >>> for ds in ts:
    ...     SlicePlot(ds, "x", ("gas", "density")).save()

    """
    # this annotation should really be Optional[Type[Dataset]]
    # but we cannot import the yt.data_objects.static_output.Dataset
    # class here without creating a circular import for now
    _dataset_cls: Optional[Type] = None

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        code_name = cls.__name__[: cls.__name__.find("Simulation")]
        if code_name:
            simulation_time_series_registry[code_name] = cls
            mylog.debug("Registering simulation: %s as %s", code_name, cls)

    def __new__(cls, outputs, *args, **kwargs):
        try:
            outputs = cls._get_filenames_from_glob_pattern(outputs)
        except TypeError:
            pass
        ret = super().__new__(cls)
        ret._pre_outputs = outputs[:]
        ret.kwargs = {}
        return ret

    def __init__(
        self,
        outputs,
        parallel=True,
        setup_function=None,
        mixed_dataset_types=False,
        **kwargs,
    ):
        # This is needed to properly set _pre_outputs for Simulation subclasses.
        self._mixed_dataset_types = mixed_dataset_types
        if is_sequence(outputs) and not isinstance(outputs, str):
            self._pre_outputs = outputs[:]
        self.tasks = AnalysisTaskProxy(self)
        self.params = TimeSeriesParametersContainer(self)
        if setup_function is None:

            def _null(x):
                return None

            setup_function = _null
        self._setup_function = setup_function
        for type_name in data_object_registry:
            setattr(
                self, type_name, functools.partial(DatasetSeriesObject, self, type_name)
            )
        self.parallel = parallel
        self.kwargs = kwargs

    @staticmethod
    def _get_filenames_from_glob_pattern(outputs):
        """
        Helper function to DatasetSeries.__new__
        handle a special case where "outputs" is assumed to be really a pattern string
        """
        pattern = outputs
        epattern = os.path.expanduser(pattern)
        data_dir = ytcfg.get("yt", "test_data_dir")
        # if no match if found from the current work dir,
        # we try to match the pattern from the test data dir
        file_list = glob.glob(epattern) or glob.glob(os.path.join(data_dir, epattern))
        if not file_list:
            raise FileNotFoundError(f"No match found for pattern : {pattern}")
        return sorted(file_list)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if isinstance(key.start, float):
                return self.get_range(key.start, key.stop)
            # This will return a sliced up object!
            return DatasetSeries(
                self._pre_outputs[key], parallel=self.parallel, **self.kwargs
            )
        o = self._pre_outputs[key]
        if isinstance(o, (str, os.PathLike)):
            o = self._load(o, **self.kwargs)
            self._setup_function(o)
        return o

    def __len__(self):
        return len(self._pre_outputs)

    @property
    def outputs(self):
        return self._pre_outputs

    def piter(self, storage=None, dynamic=False):
        r"""Iterate over time series components in parallel.

        This allows you to iterate over a time series while dispatching
        individual components of that time series to different processors or
        processor groups.  If the parallelism strategy was set to be
        multi-processor (by "parallel = N" where N is an integer when the
        DatasetSeries was created) this will issue each dataset to an
        N-processor group.  For instance, this would allow you to start a 1024
        processor job, loading up 100 datasets in a time series and creating 8
        processor groups of 128 processors each, each of which would be
        assigned a different dataset.  This could be accomplished as shown in
        the examples below.  The *storage* option is as seen in
        :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_objects`
        which is a mechanism for storing results of analysis on an individual
        dataset and then combining the results at the end, so that the entire
        set of processors have access to those results.

        Note that supplying a *store* changes the iteration mechanism; see
        below.

        Parameters
        ----------
        storage : dict
            This is a dictionary, which will be filled with results during the
            course of the iteration.  The keys will be the dataset
            indices and the values will be whatever is assigned to the *result*
            attribute on the storage during iteration.
        dynamic : boolean
            This governs whether or not dynamic load balancing will be
            enabled.  This requires one dedicated processor; if this
            is enabled with a set of 128 processors available, only
            127 will be available to iterate over objects as one will
            be load balancing the rest.


        Examples
        --------
        Here is an example of iteration when the results do not need to be
        stored.  One processor will be assigned to each dataset.

        >>> ts = DatasetSeries("DD*/DD*.index")
        >>> for ds in ts.piter():
        ...     SlicePlot(ds, "x", ("gas", "density")).save()
        ...

        This demonstrates how one might store results:

        >>> def print_time(ds):
        ...     print(ds.current_time)
        ...
        >>> ts = DatasetSeries("DD*/DD*.index", setup_function=print_time)
        ...
        >>> my_storage = {}
        >>> for sto, ds in ts.piter(storage=my_storage):
        ...     v, c = ds.find_max(("gas", "density"))
        ...     sto.result = (v, c)
        ...
        >>> for i, (v, c) in sorted(my_storage.items()):
        ...     print("% 4i  %0.3e" % (i, v))
        ...

        This shows how to dispatch 4 processors to each dataset:

        >>> ts = DatasetSeries("DD*/DD*.index", parallel=4)
        >>> for ds in ts.piter():
        ...     ProjectionPlot(ds, "x", ("gas", "density")).save()
        ...

        """
        if not self.parallel:
            njobs = 1
        elif not dynamic:
            if self.parallel:
                njobs = -1
            else:
                njobs = self.parallel
        else:
            my_communicator = communication_system.communicators[-1]
            nsize = my_communicator.size
            if nsize == 1:
                self.parallel = False
                dynamic = False
                njobs = 1
            else:
                njobs = nsize - 1

        for output in parallel_objects(
            self._pre_outputs, njobs=njobs, storage=storage, dynamic=dynamic
        ):
            if storage is not None:
                sto, output = output

            if isinstance(output, str):
                ds = self._load(output, **self.kwargs)
                self._setup_function(ds)
            else:
                ds = output

            if storage is not None:
                next_ret = (sto, ds)
            else:
                next_ret = ds

            yield next_ret

    def eval(self, tasks, obj=None):
        return_values = {}
        for store, ds in self.piter(return_values):
            store.result = []
            for task in always_iterable(tasks):
                try:
                    style = inspect.getargspec(task.eval)[0][1]
                    if style == "ds":
                        arg = ds
                    elif style == "data_object":
                        if obj is None:
                            obj = DatasetSeriesObject(self, "all_data")
                        arg = obj.get(ds)
                    rv = task.eval(arg)
                # We catch and store YT-originating exceptions
                # This fixes the standard problem of having a sphere that's too
                # small.
                except YTException:
                    pass
                store.result.append(rv)
        return [v for k, v in sorted(return_values.items())]

    @classmethod
    def from_output_log(cls, output_log, line_prefix="DATASET WRITTEN", parallel=True):
        filenames = []
        for line in open(output_log):
            if not line.startswith(line_prefix):
                continue
            cut_line = line[len(line_prefix) :].strip()
            fn = cut_line.split()[0]
            filenames.append(fn)
        obj = cls(filenames, parallel=parallel)
        return obj

    def _load(self, output_fn, *, hint: Optional[str] = None, **kwargs):
        from yt.loaders import load

        if self._dataset_cls is not None:
            return self._dataset_cls(output_fn, **kwargs)
        elif self._mixed_dataset_types:
            return load(output_fn, hint=hint, **kwargs)
        ds = load(output_fn, hint=hint, **kwargs)
        self._dataset_cls = ds.__class__
        return ds

    def particle_trajectories(
        self, indices, fields=None, suppress_logging=False, ptype=None
    ):
        r"""Create a collection of particle trajectories in time over a series of
        datasets.

        Parameters
        ----------
        indices : array_like
            An integer array of particle indices whose trajectories we
            want to track. If they are not sorted they will be sorted.
        fields : list of strings, optional
            A set of fields that is retrieved when the trajectory
            collection is instantiated. Default: None (will default
            to the fields 'particle_position_x', 'particle_position_y',
            'particle_position_z')
        suppress_logging : boolean
            Suppress yt's logging when iterating over the simulation time
            series. Default: False
        ptype : str, optional
            Only use this particle type. Default: None, which uses all particle type.

        Examples
        --------
        >>> my_fns = glob.glob("orbit_hdf5_chk_00[0-9][0-9]")
        >>> my_fns.sort()
        >>> fields = [
        ...     ("all", "particle_position_x"),
        ...     ("all", "particle_position_y"),
        ...     ("all", "particle_position_z"),
        ...     ("all", "particle_velocity_x"),
        ...     ("all", "particle_velocity_y"),
        ...     ("all", "particle_velocity_z"),
        ... ]
        >>> ds = load(my_fns[0])
        >>> init_sphere = ds.sphere(ds.domain_center, (0.5, "unitary"))
        >>> indices = init_sphere[("all", "particle_index")].astype("int")
        >>> ts = DatasetSeries(my_fns)
        >>> trajs = ts.particle_trajectories(indices, fields=fields)
        >>> for t in trajs:
        ...     print(
        ...         t[("all", "particle_velocity_x")].max(),
        ...         t[("all", "particle_velocity_x")].min(),
        ...     )

        Notes
        -----
        This function will fail if there are duplicate particle ids or if some of the
        particle disappear.
        """
        return ParticleTrajectories(
            self, indices, fields=fields, suppress_logging=suppress_logging, ptype=ptype
        )


class TimeSeriesQuantitiesContainer:
    def __init__(self, data_object, quantities):
        self.data_object = data_object
        self.quantities = quantities

    def __getitem__(self, key):
        if key not in self.quantities:
            raise KeyError(key)
        q = self.quantities[key]

        def run_quantity_wrapper(quantity, quantity_name):
            @wraps(derived_quantity_registry[quantity_name][1])
            def run_quantity(*args, **kwargs):
                to_run = quantity(*args, **kwargs)
                return self.data_object.eval(to_run)

            return run_quantity

        return run_quantity_wrapper(q, key)


class DatasetSeriesObject:
    def __init__(self, time_series, data_object_name, *args, **kwargs):
        self.time_series = weakref.proxy(time_series)
        self.data_object_name = data_object_name
        self._args = args
        self._kwargs = kwargs
        qs = {
            qn: create_quantity_proxy(qv)
            for qn, qv in derived_quantity_registry.items()
        }
        self.quantities = TimeSeriesQuantitiesContainer(self, qs)

    def eval(self, tasks):
        return self.time_series.eval(tasks, self)

    def get(self, ds):
        # We get the type name, which corresponds to an attribute of the
        # index
        cls = getattr(ds, self.data_object_name)
        return cls(*self._args, **self._kwargs)


class SimulationTimeSeries(DatasetSeries):
    def __init__(self, parameter_filename, find_outputs=False):
        """
        Base class for generating simulation time series types.
        Principally consists of a *parameter_filename*.
        """

        if not os.path.exists(parameter_filename):
            raise FileNotFoundError(parameter_filename)
        self.parameter_filename = parameter_filename
        self.basename = os.path.basename(parameter_filename)
        self.directory = os.path.dirname(parameter_filename)
        self.parameters = {}
        self.key_parameters = []

        # Set some parameter defaults.
        self._set_parameter_defaults()
        # Read the simulation dataset.
        self._parse_parameter_file()
        # Set units
        self._set_units()
        # Figure out the starting and stopping times and redshift.
        self._calculate_simulation_bounds()
        # Get all possible datasets.
        self._get_all_outputs(find_outputs=find_outputs)

        self.print_key_parameters()

    def _set_parameter_defaults(self):
        pass

    def _parse_parameter_file(self):
        pass

    def _set_units(self):
        pass

    def _calculate_simulation_bounds(self):
        pass

    def _get_all_outputs(**kwargs):
        pass

    def __repr__(self):
        return self.parameter_filename

    _arr = None

    @property
    def arr(self):
        if self._arr is not None:
            return self._arr
        self._arr = functools.partial(YTArray, registry=self.unit_registry)
        return self._arr

    _quan = None

    @property
    def quan(self):
        if self._quan is not None:
            return self._quan
        self._quan = functools.partial(YTQuantity, registry=self.unit_registry)
        return self._quan

    @parallel_root_only
    def print_key_parameters(self):
        """
        Print out some key parameters for the simulation.
        """
        if self.simulation_type == "grid":
            for a in ["domain_dimensions", "domain_left_edge", "domain_right_edge"]:
                self._print_attr(a)
        for a in ["initial_time", "final_time", "cosmological_simulation"]:
            self._print_attr(a)
        if getattr(self, "cosmological_simulation", False):
            for a in [
                "box_size",
                "omega_matter",
                "omega_lambda",
                "omega_radiation",
                "hubble_constant",
                "initial_redshift",
                "final_redshift",
            ]:
                self._print_attr(a)
        for a in self.key_parameters:
            self._print_attr(a)
        mylog.info("Total datasets: %d.", len(self.all_outputs))

    def _print_attr(self, a):
        """
        Print the attribute or warn about it missing.
        """
        if not hasattr(self, a):
            mylog.error("Missing %s in dataset definition!", a)
            return
        v = getattr(self, a)
        mylog.info("Parameters: %-25s = %s", a, v)

    def _get_outputs_by_key(self, key, values, tolerance=None, outputs=None):
        r"""
        Get datasets at or near to given values.

        Parameters
        ----------
        key : str
            The key by which to retrieve outputs, usually 'time' or
            'redshift'.
        values : array_like
            A list of values, given as floats.
        tolerance : float
            If not None, do not return a dataset unless the value is
            within the tolerance value.  If None, simply return the
            nearest dataset.
            Default: None.
        outputs : list
            The list of outputs from which to choose.  If None,
            self.all_outputs is used.
            Default: None.

        Examples
        --------
        >>> datasets = es.get_outputs_by_key("redshift", [0, 1, 2], tolerance=0.1)

        """

        if not isinstance(values, YTArray):
            if isinstance(values, tuple) and len(values) == 2:
                values = self.arr(*values)
            else:
                values = self.arr(values)
        values = values.in_base()

        if outputs is None:
            outputs = self.all_outputs
        my_outputs = []
        if not outputs:
            return my_outputs
        for value in values:
            outputs.sort(key=lambda obj, value=value: np.abs(value - obj[key]))
            if (
                tolerance is None or np.abs(value - outputs[0][key]) <= tolerance
            ) and outputs[0] not in my_outputs:
                my_outputs.append(outputs[0])
            else:
                mylog.error("No dataset added for %s = %f.", key, value)

        outputs.sort(key=lambda obj: obj["time"])
        return my_outputs



from . import construction_data_containers as __cdc, selection_objects as __sdc
from .analyzer_objects import AnalysisTask, analysis_task
from .image_array import ImageArray
from .index_subobjects import AMRGridPatch, OctreeSubset
from .particle_filters import add_particle_filter, particle_filter
from .profiles import ParticleProfile, Profile1D, Profile2D, Profile3D, create_profile
from .static_output import Dataset
from .time_series import DatasetSeries, DatasetSeriesObject



r"""Routines for tracking the scaling of the triangulation routines."""
import cProfile
import os
import pstats
import time

import numpy as np

from yt.utilities.lib.cykdtree.tests import run_test


def stats_run(
    npart,
    nproc,
    ndim,
    periodic=False,
    overwrite=False,
    display=False,
    suppress_final_output=False,
):
    r"""Get timing stats using :package:`cProfile`.

    Args:
        npart (int): Number of particles.
        nproc (int): Number of processors.
        ndim (int): Number of dimensions.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.
        display (bool, optional): If True, display the profile results.
            Defaults to False.

    """
    perstr = ""
    outstr = ""
    if periodic:
        perstr = "_periodic"
    if suppress_final_output:
        outstr = "_noout"
    fname_stat = f"stat_{npart}part_{nproc}proc_{ndim}dim{perstr}{outstr}.txt"
    if overwrite or not os.path.isfile(fname_stat):
        cProfile.run(
            "from yt.utilities.lib.cykdtree.tests import run_test; "
            + f"run_test({npart}, {ndim}, nproc={nproc}, "
            + f"periodic={periodic}, "
            + f"suppress_final_output={suppress_final_output})",
            fname_stat,
        )
    if display:
        p = pstats.Stats(fname_stat)
        p.sort_stats("time").print_stats(10)
        return p
    return fname_stat


def time_run(
    npart, nproc, ndim, nrep=1, periodic=False, leafsize=10, suppress_final_output=False
):
    r"""Get running times using :package:`time`.

    Args:
        npart (int): Number of particles.
        nproc (int): Number of processors.
        ndim (int): Number of dimensions.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        leafsize (int, optional): The maximum number of points that should be
            in any leaf in the tree. Defaults to 10.
        suppress_final_output (bool, optional): If True, the final output
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.

    """
    times = np.empty(nrep, "float")
    for i in range(nrep):
        t1 = time.time()
        run_test(
            npart,
            ndim,
            nproc=nproc,
            periodic=periodic,
            leafsize=leafsize,
            suppress_final_output=suppress_final_output,
        )
        t2 = time.time()
        times[i] = t2 - t1
    return np.mean(times), np.std(times)


def strong_scaling(
    npart=1e6,
    nrep=1,
    periodic=False,
    leafsize=10,
    overwrite=True,
    suppress_final_output=False,
):
    r"""Plot the scaling with number of processors for a particular function.

    Args:
        npart (int, optional): Number of particles. Defaults to 1e6.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        leafsize (int, optional): The maximum number of points that should be
            in any leaf in the tree. Defaults to 10.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.

    """
    import matplotlib.pyplot as plt

    npart = int(npart)
    perstr = ""
    outstr = ""
    if periodic:
        perstr = "_periodic"
    if suppress_final_output:
        outstr = "_noout"
    fname_plot = "plot_strong_scaling_nproc_{}part{}_{}leafsize{}.png".format(
        npart, perstr, leafsize, outstr
    )
    nproc_list = [1, 2, 4, 8]  # , 16]
    ndim_list = [2, 3, 4]
    clr_list = ["b", "r", "g", "m"]
    times = np.empty((len(nproc_list), len(ndim_list), 2), "float")
    for j, nproc in enumerate(nproc_list):
        for i, ndim in enumerate(ndim_list):
            times[j, i, 0], times[j, i, 1] = time_run(
                npart,
                nproc,
                ndim,
                nrep=nrep,
                periodic=periodic,
                leafsize=leafsize,
                suppress_final_output=suppress_final_output,
            )
            print(f"Finished {ndim}D on {nproc}.")
    fig, axs = plt.subplots(1, 1)
    for i in range(len(ndim_list)):
        ndim = ndim_list[i]
        clr = clr_list[i]
        axs.errorbar(
            nproc_list,
            times[:, i, 0],
            yerr=times[:, i, 1],
            fmt=clr,
            label=f"ndim = {ndim}",
        )
    axs.set_xlabel("# of Processors")
    axs.set_ylabel("Time (s)")
    axs.legend()
    fig.savefig(fname_plot)
    print("    " + fname_plot)


def weak_scaling(
    npart=1e4,
    nrep=1,
    periodic=False,
    leafsize=10,
    overwrite=True,
    suppress_final_output=False,
):
    r"""Plot the scaling with number of processors with a constant number of
    particles per processor for a particular function.

    Args:
        npart (int, optional): Number of particles per processor. Defaults to
            1e4.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        leafsize (int, optional): The maximum number of points that should be
            in any leaf in the tree. Defaults to 10.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.

    """
    import matplotlib.pyplot as plt

    npart = int(npart)
    perstr = ""
    outstr = ""
    if periodic:
        perstr = "_periodic"
    if suppress_final_output:
        outstr = "_noout"
    fname_plot = "plot_weak_scaling_nproc_{}part{}_{}leafsize{}.png".format(
        npart, perstr, leafsize, outstr
    )
    nproc_list = [1, 2, 4, 8, 16]
    ndim_list = [2, 3]
    clr_list = ["b", "r", "g", "m"]
    times = np.empty((len(nproc_list), len(ndim_list), 2), "float")
    for j, nproc in enumerate(nproc_list):
        for i, ndim in enumerate(ndim_list):
            times[j, i, 0], times[j, i, 1] = time_run(
                npart * nproc,
                nproc,
                ndim,
                nrep=nrep,
                periodic=periodic,
                leafsize=leafsize,
                suppress_final_output=suppress_final_output,
            )
    fig, axs = plt.subplots(1, 1)
    for i in range(len(ndim_list)):
        ndim = ndim_list[i]
        clr = clr_list[i]
        axs.errorbar(
            nproc_list,
            times[:, i, 0],
            yerr=times[:, i, 1],
            fmt=clr,
            label=f"ndim = {ndim}",
        )
    axs.set_xlabel("# of Processors")
    axs.set_ylabel("Time (s)")
    axs.legend()
    fig.savefig(fname_plot)
    print("    " + fname_plot)



from typing import List, Tuple

import numpy as np


class ZBuffer:
    """A container object for z-buffer arrays

    A zbuffer is a companion array for an image that allows the volume rendering
    infrastructure to determine whether one opaque source is in front of another
    opaque source.  The z buffer encodes the distance to the opaque source
    relative to the camera position.

    Parameters
    ----------
    rgba: MxNx4 image
        The image the z buffer corresponds to
    z: MxN image
        The z depth of each pixel in the image. The shape of the image must be
        the same as each RGBA channel in the original image.

    Examples
    --------
    >>> import numpy as np
    >>> shape = (64, 64)
    >>> b1 = Zbuffer(np.random.random(shape), np.ones(shape))
    >>> b2 = Zbuffer(np.random.random(shape), np.zeros(shape))
    >>> c = b1 + b2
    >>> np.all(c.rgba == b2.rgba)
    True
    >>> np.all(c.z == b2.z)
    True
    >>> np.all(c == b2)
    True

    """

    def __init__(self, rgba, z):
        super().__init__()
        assert rgba.shape[: len(z.shape)] == z.shape
        self.rgba = rgba
        self.z = z
        self.shape = z.shape

    def __add__(self, other):
        assert self.shape == other.shape
        f = self.z < other.z
        if self.z.shape[1] == 1:
            # Non-rectangular
            rgba = self.rgba * f[:, None, :]
            rgba += other.rgba * (1.0 - f)[:, None, :]
        else:
            b = self.z > other.z
            rgba = np.zeros(self.rgba.shape)
            rgba[f] = self.rgba[f]
            rgba[b] = other.rgba[b]
        z = np.min([self.z, other.z], axis=0)
        return ZBuffer(rgba, z)

    def __iadd__(self, other):
        tmp = self + other
        self.rgba = tmp.rgba
        self.z = tmp.z
        return self

    def __eq__(self, other):
        equal = True
        equal *= np.all(self.rgba == other.rgba)
        equal *= np.all(self.z == other.z)
        return equal

    def paint(self, ind, value, z):
        if z < self.z[ind]:
            self.rgba[ind] = value
            self.z[ind] = z


if __name__ == "__main__":
    shape: Tuple[int, ...] = (64, 64)
    shapes: List[Tuple[int, ...]] = [(64, 64), (16, 16, 4), (128,), (16, 32)]
    for shape in shapes:
        b1 = ZBuffer(np.random.random(shape), np.ones(shape))
        b2 = ZBuffer(np.random.random(shape), np.zeros(shape))
        c = b1 + b2
        assert np.all(c.rgba == b2.rgba)
        assert np.all(c.z == b2.z)
        assert np.all(c == b2)



import numpy.testing as npt

from yt.testing import fake_random_ds
from yt.visualization.api import SlicePlot


def test_sliceplot_set_unit_and_zlim_order():
    ds = fake_random_ds(16)
    field = ("gas", "density")

    p0 = SlicePlot(ds, "z", field)
    p0.set_unit(field, "kg/m**3")
    p0.set_zlim(field, zmin=0)

    # reversing order of operations
    p1 = SlicePlot(ds, "z", field)
    p1.set_zlim(field, zmin=0)
    p1.set_unit(field, "kg/m**3")

    p0._setup_plots()
    p1._setup_plots()

    im0 = p0.plots[field].image._A
    im1 = p1.plots[field].image._A

    npt.assert_allclose(im0, im1)



import yt
from yt.testing import fake_amr_ds, requires_external_executable, requires_module


@requires_external_executable("tex")
@requires_module("pyx")
def test_eps_writer(tmp_path):
    import yt.visualization.eps_writer as eps

    fields = [
        ("gas", "density"),
        ("gas", "temperature"),
    ]
    units = [
        "g/cm**3",
        "K",
    ]
    ds = fake_amr_ds(fields=fields, units=units)
    slc = yt.SlicePlot(
        ds,
        "z",
        fields=fields,
    )
    eps_fig = eps.multiplot_yt(2, 1, slc, bare_axes=True)
    eps_fig.scale_line(0.2, "5 cm")
    savefile = tmp_path / "multi"
    eps_fig.save_fig(savefile, format="eps")
    assert savefile.with_suffix(".eps").exists()



from yt.frontends.eagle.definitions import eaglenetwork_ion_lookup
from yt.frontends.owls.fields import OWLSFieldInfo
from yt.units.yt_array import YTQuantity
from yt.utilities.periodic_table import periodic_table


class EagleNetworkFieldInfo(OWLSFieldInfo):

    _ions = (
        "H1",
        "H2",
        "He1",
        "He2",
        "He3",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "C7",
        "N1",
        "N2",
        "N3",
        "N4",
        "N5",
        "N6",
        "N7",
        "N8",
        "O1",
        "O2",
        "O3",
        "O4",
        "O5",
        "O6",
        "O7",
        "O8",
        "O9",
        "Ne1",
        "Ne2",
        "Ne3",
        "Ne4",
        "Ne5",
        "Ne6",
        "Ne7",
        "Ne8",
        "Ne9",
        "Ne10",
        "Ne11",
        "Mg1",
        "Mg2",
        "Mg3",
        "Mg4",
        "Mg5",
        "Mg6",
        "Mg7",
        "Mg8",
        "Mg9",
        "Mg10",
        "Mg11",
        "Mg12",
        "Mg13",
        "Si1",
        "Si2",
        "Si3",
        "Si4",
        "Si5",
        "Si6",
        "Si7",
        "Si8",
        "Si9",
        "Si10",
        "Si11",
        "Si12",
        "Si13",
        "Si14",
        "Si15",
        "Si16",
        "Si17",
        "Ca1",
        "Ca2",
        "Ca3",
        "Ca4",
        "Ca5",
        "Ca6",
        "Ca7",
        "Ca8",
        "Ca9",
        "Ca10",
        "Ca11",
        "Ca12",
        "Ca13",
        "Ca14",
        "Ca15",
        "Ca16",
        "Ca17",
        "Ca18",
        "Ca19",
        "Ca20",
        "Ca21",
        "Fe1",
        "Fe2",
        "Fe3",
        "Fe4",
        "Fe5",
        "Fe6",
        "Fe7",
        "Fe8",
        "Fe9",
        "Fe10",
        "Fe11",
        "Fe12",
        "Fe13",
        "Fe14",
        "Fe15",
        "Fe16",
        "Fe17",
        "Fe18",
        "Fe19",
        "Fe20",
        "Fe21",
        "Fe22",
        "Fe23",
        "Fe24",
        "Fe25",
        "Fe25",
        "Fe27",
    )

    def __init__(self, ds, field_list, slice_info=None):
        super().__init__(ds, field_list, slice_info=slice_info)

    def _create_ion_density_func(self, ftype, ion):
        """returns a function that calculates the ion density of a particle."""

        def _ion_density(field, data):

            # Lookup the index of the ion
            index = eaglenetwork_ion_lookup[ion]

            # Ion to hydrogen number density ratio
            ion_chem = data[ftype, "Chemistry_%03i" % index]

            # Mass of a single ion
            if ion[0:2].isalpha():
                symbol = ion[0:2].capitalize()
            else:
                symbol = ion[0:1].capitalize()
            m_ion = YTQuantity(periodic_table.elements_by_symbol[symbol].weight, "amu")

            # hydrogen number density
            n_H = data["PartType0", "H_number_density"]

            return m_ion * ion_chem * n_H

        return _ion_density



from unyt.unit_systems import *


def create_code_unit_system(unit_registry, current_mks_unit=None):
    code_unit_system = UnitSystem(
        name=unit_registry.unit_system_id,
        length_unit="code_length",
        mass_unit="code_mass",
        time_unit="code_time",
        temperature_unit="code_temperature",
        current_mks_unit=current_mks_unit,
        registry=unit_registry,
    )
    code_unit_system["velocity"] = "code_velocity"
    if current_mks_unit:
        code_unit_system["magnetic_field_mks"] = "code_magnetic"
    else:
        code_unit_system["magnetic_field_cgs"] = "code_magnetic"
    code_unit_system["pressure"] = "code_pressure"
    return code_unit_system



import glob
import os

import numpy as np

from yt.data_objects.static_output import Dataset
from yt.frontends.halo_catalog.data_structures import HaloCatalogFile
from yt.funcs import setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.cosmology import Cosmology

from .fields import AHFHalosFieldInfo


class AHFHalosFile(HaloCatalogFile):
    def __init__(self, ds, io, filename, file_id, range=None):
        root, _ = os.path.splitext(filename)
        candidates = glob.glob(root + "*.AHF_halos")
        if len(candidates) == 1:
            filename = candidates[0]
        else:
            raise ValueError("Too many AHF_halos files.")
        self.col_names = self._read_column_names(filename)
        super().__init__(ds, io, filename, file_id, range)

    def read_data(self, usecols=None):
        return np.genfromtxt(self.filename, names=self.col_names, usecols=usecols)

    def _read_column_names(self, filename):
        with open(filename) as f:
            line = f.readline()
            # Remove leading '#'
            line = line[1:]
            names = line.split()
            # Remove trailing '()'
            names = [name.split("(")[0] for name in names]
            return names

    def _read_particle_positions(self, ptype, f=None):
        """
        Read all particle positions in this file.
        """

        halos = self.read_data(usecols=["Xc", "Yc", "Zc"])
        pos = np.empty((halos.size, 3), dtype="float64")
        for i, ax in enumerate("XYZ"):
            pos[:, i] = halos[f"{ax}c"].astype("float64")

        return pos


class AHFHalosDataset(Dataset):
    _index_class = ParticleIndex
    _file_class = AHFHalosFile
    _field_info_class = AHFHalosFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="ahf",
        n_ref=16,
        num_zones=2,
        units_override=None,
        unit_system="cgs",
        hubble_constant=1.0,
    ):
        root, _ = os.path.splitext(filename)
        self.log_filename = root + ".log"
        self.hubble_constant = hubble_constant

        self.n_ref = n_ref
        self.num_zones = num_zones
        super().__init__(
            filename,
            dataset_type=dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _set_code_unit_attributes(self):
        setdefaultattr(self, "length_unit", self.quan(1.0, "kpccm/h"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "Msun/h"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        setdefaultattr(self, "velocity_unit", self.quan(1.0, "km/s"))

    def _parse_parameter_file(self):
        # Read all parameters.
        simu = self._read_log_simu()
        param = self._read_parameter()

        # Set up general information.
        self.filename_template = self.parameter_filename
        self.file_count = 1
        self.parameters.update(param)
        self.particle_types = "halos"
        self.particle_types_raw = "halos"

        # Set up geometrical information.
        self.refine_by = 2
        self.dimensionality = 3
        nz = self.num_zones
        self.domain_dimensions = np.ones(self.dimensionality, "int32") * nz
        self.domain_left_edge = np.array([0.0, 0.0, 0.0])
        # Note that boxsize is in Mpc but particle positions are in kpc.
        self.domain_right_edge = np.array([simu["boxsize"]] * 3) * 1000
        self._periodicity = (True, True, True)

        # Set up cosmological information.
        self.cosmological_simulation = 1
        self.current_redshift = param["z"]
        self.omega_lambda = simu["lambda0"]
        self.omega_matter = simu["omega0"]
        cosmo = Cosmology(
            hubble_constant=self.hubble_constant,
            omega_matter=self.omega_matter,
            omega_lambda=self.omega_lambda,
        )
        self.current_time = cosmo.lookback_time(param["z"], 1e6).in_units("s")

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".parameter"):
            return False
        with open(filename) as f:
            if f.readlines()[11].startswith("AHF"):
                return True
        return False

    # Helper methods

    def _read_log_simu(self):
        simu = {}
        with open(self.log_filename) as f:
            for l in f:
                if l.startswith("simu."):
                    name, val = l.split(":")
                    key = name.strip().split(".")[1]
                    try:
                        val = float(val)
                    except Exception:
                        val = float.fromhex(val)
                    simu[key] = val
        return simu

    def _read_parameter(self):
        param = {}
        with open(self.parameter_filename) as f:
            for l in f:
                words = l.split()
                if len(words) == 2:
                    key, val = words
                    try:
                        val = float(val)
                        param[key] = val
                    except Exception:
                        pass
        return param

    @property
    def _skip_cache(self):
        return True



"""
Title: framework.py
Purpose: Contains answer tests that are used by yt's various frontends
"""
import contextlib
import glob
import hashlib
import logging
import os
import pickle
import shelve
import sys
import tempfile
import time
import urllib
import warnings
import zlib
from collections import defaultdict
from typing import Optional

import numpy as np
from matplotlib import image as mpimg
from matplotlib.testing.compare import compare_images
from nose.plugins import Plugin

from yt.config import ytcfg
from yt.data_objects.static_output import Dataset
from yt.funcs import get_pbar, get_yt_version
from yt.loaders import load, load_simulation
from yt.testing import (
    assert_allclose_units,
    assert_almost_equal,
    assert_equal,
    assert_rel_equal,
    skipif,
)
from yt.utilities.exceptions import (
    YTAmbiguousDataType,
    YTCloudError,
    YTNoAnswerNameSpecified,
    YTNoOldAnswer,
    YTUnidentifiedDataType,
)
from yt.utilities.logger import disable_stream_logging
from yt.visualization import (
    image_writer as image_writer,
    particle_plots as particle_plots,
    plot_window as pw,
    profile_plotter as profile_plotter,
)

mylog = logging.getLogger("nose.plugins.answer-testing")
run_big_data = False

# Set the latest gold and local standard filenames
_latest = ytcfg.get("yt", "gold_standard_filename")
_latest_local = ytcfg.get("yt", "local_standard_filename")
_url_path = ytcfg.get("yt", "answer_tests_url")


class AnswerTesting(Plugin):
    name = "answer-testing"
    _my_version = None

    def options(self, parser, env=os.environ):
        super().options(parser, env=env)
        parser.add_option(
            "--answer-name",
            dest="answer_name",
            metavar="str",
            default=None,
            help="The name of the standard to store/compare against",
        )
        parser.add_option(
            "--answer-store",
            dest="store_results",
            metavar="bool",
            default=False,
            action="store_true",
            help="Should we store this result instead of comparing?",
        )
        parser.add_option(
            "--local",
            dest="local_results",
            default=False,
            action="store_true",
            help="Store/load reference results locally?",
        )
        parser.add_option(
            "--answer-big-data",
            dest="big_data",
            default=False,
            help="Should we run against big data, too?",
            action="store_true",
        )
        parser.add_option(
            "--local-dir",
            dest="output_dir",
            metavar="str",
            help="The name of the directory to store local results",
        )

    @property
    def my_version(self, version=None):
        if self._my_version is not None:
            return self._my_version
        if version is None:
            try:
                version = get_yt_version()
            except Exception:
                version = f"UNKNOWN{time.time()}"
        self._my_version = version
        return self._my_version

    def configure(self, options, conf):
        super().configure(options, conf)
        if not self.enabled:
            return
        disable_stream_logging()

        # Parse through the storage flags to make sense of them
        # and use reasonable defaults
        # If we're storing the data, default storage name is local
        # latest version
        if options.store_results:
            if options.answer_name is None:
                self.store_name = _latest_local
            else:
                self.store_name = options.answer_name
            self.compare_name = None
        # if we're not storing, then we're comparing, and we want default
        # comparison name to be the latest gold standard
        # either on network or local
        else:
            if options.answer_name is None:
                if options.local_results:
                    self.compare_name = _latest_local
                else:
                    self.compare_name = _latest
            else:
                self.compare_name = options.answer_name
            self.store_name = self.my_version

        self.store_results = options.store_results

        ytcfg["yt", "internals", "within_testing"] = True
        AnswerTestingTest.result_storage = self.result_storage = defaultdict(dict)
        if self.compare_name == "SKIP":
            self.compare_name = None
        elif self.compare_name == "latest":
            self.compare_name = _latest

        # Local/Cloud storage
        if options.local_results:
            if options.output_dir is None:
                print("Please supply an output directory with the --local-dir option")
                sys.exit(1)
            storage_class = AnswerTestLocalStorage
            output_dir = os.path.realpath(options.output_dir)
            # Fix up filename for local storage
            if self.compare_name is not None:
                self.compare_name = os.path.join(
                    output_dir, self.compare_name, self.compare_name
                )

            # Create a local directory only when `options.answer_name` is
            # provided. If it is not provided then creating local directory
            # will depend on the `AnswerTestingTest.answer_name` value of the
            # test, this case is handled in AnswerTestingTest class.
            if options.store_results and options.answer_name is not None:
                name_dir_path = os.path.join(output_dir, self.store_name)
                if not os.path.isdir(name_dir_path):
                    os.makedirs(name_dir_path)
                self.store_name = os.path.join(name_dir_path, self.store_name)
        else:
            storage_class = AnswerTestCloudStorage

        # Initialize answer/reference storage
        AnswerTestingTest.reference_storage = self.storage = storage_class(
            self.compare_name, self.store_name
        )
        AnswerTestingTest.options = options

        self.local_results = options.local_results
        global run_big_data
        run_big_data = options.big_data

    def finalize(self, result=None):
        if not self.store_results:
            return
        self.storage.dump(self.result_storage)

    def help(self):
        return "yt answer testing support"


class AnswerTestStorage:
    def __init__(self, reference_name=None, answer_name=None):
        self.reference_name = reference_name
        self.answer_name = answer_name
        self.cache = {}

    def dump(self, result_storage, result):
        raise NotImplementedError

    def get(self, ds_name, default=None):
        raise NotImplementedError


class AnswerTestCloudStorage(AnswerTestStorage):
    def get(self, ds_name, default=None):
        if self.reference_name is None:
            return default
        if ds_name in self.cache:
            return self.cache[ds_name]
        url = _url_path.format(self.reference_name, ds_name)
        try:
            resp = urllib.request.urlopen(url)
        except urllib.error.HTTPError:
            raise YTNoOldAnswer(url)
        else:
            for _ in range(3):
                try:
                    data = resp.read()
                except Exception:
                    time.sleep(0.01)
                else:
                    # We were successful
                    break
            else:
                # Raise error if all tries were unsuccessful
                raise YTCloudError(url)
            # This is dangerous, but we have a controlled S3 environment
            rv = pickle.loads(data)
        self.cache[ds_name] = rv
        return rv

    def progress_callback(self, current, total):
        self.pbar.update(current)

    def dump(self, result_storage):
        if self.answer_name is None:
            return
        # This is where we dump our result storage up to Amazon, if we are able
        # to.
        import pyrax

        credentials = os.path.expanduser(os.path.join("~", ".yt", "rackspace"))
        pyrax.set_credential_file(credentials)
        cf = pyrax.cloudfiles
        c = cf.get_container("yt-answer-tests")
        pb = get_pbar("Storing results ", len(result_storage))
        for i, ds_name in enumerate(result_storage):
            pb.update(i + 1)
            rs = pickle.dumps(result_storage[ds_name])
            object_name = f"{self.answer_name}_{ds_name}"
            if object_name in c.get_object_names():
                obj = c.get_object(object_name)
                c.delete_object(obj)
            c.store_object(object_name, rs)
        pb.finish()


class AnswerTestLocalStorage(AnswerTestStorage):
    def dump(self, result_storage):
        # The 'tainted' attribute is automatically set to 'True'
        # if the dataset required for an answer test is missing
        # (see can_run_ds().
        # This logic check prevents creating a shelve with empty answers.
        storage_is_tainted = result_storage.get("tainted", False)
        if self.answer_name is None or storage_is_tainted:
            return
        # Store data using shelve
        ds = shelve.open(self.answer_name, protocol=-1)
        for ds_name in result_storage:
            answer_name = f"{ds_name}"
            if answer_name in ds:
                mylog.info("Overwriting %s", answer_name)
            ds[answer_name] = result_storage[ds_name]
        ds.close()

    def get(self, ds_name, default=None):
        if self.reference_name is None:
            return default
        # Read data using shelve
        answer_name = f"{ds_name}"
        os.makedirs(os.path.dirname(self.reference_name), exist_ok=True)
        ds = shelve.open(self.reference_name, protocol=-1)
        try:
            result = ds[answer_name]
        except KeyError:
            result = default
        ds.close()
        return result


@contextlib.contextmanager
def temp_cwd(cwd):
    oldcwd = os.getcwd()
    os.chdir(cwd)
    yield
    os.chdir(oldcwd)


def can_run_ds(ds_fn, file_check=False):
    result_storage = AnswerTestingTest.result_storage
    if isinstance(ds_fn, Dataset):
        return result_storage is not None
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, ds_fn)) and result_storage is not None
    try:
        load(ds_fn)
    except FileNotFoundError:
        if ytcfg.get("yt", "internals", "strict_requires"):
            if result_storage is not None:
                result_storage["tainted"] = True
            raise
        return False
    except (YTUnidentifiedDataType, YTAmbiguousDataType):
        return False

    return result_storage is not None


def data_dir_load(ds_fn, cls=None, args=None, kwargs=None):
    args = args or ()
    kwargs = kwargs or {}
    path = ytcfg.get("yt", "test_data_dir")
    if isinstance(ds_fn, Dataset):
        return ds_fn
    if not os.path.isdir(path):
        return False
    if cls is None:
        ds = load(ds_fn, *args, **kwargs)
    else:
        ds = cls(os.path.join(path, ds_fn), *args, **kwargs)
    ds.index
    return ds


def sim_dir_load(sim_fn, path=None, sim_type="Enzo", find_outputs=False):
    if path is None and not os.path.exists(sim_fn):
        raise OSError
    if os.path.exists(sim_fn) or not path:
        path = "."
    return load_simulation(
        os.path.join(path, sim_fn), sim_type, find_outputs=find_outputs
    )


class AnswerTestingTest:
    reference_storage = None
    result_storage = None
    prefix = ""
    options = None
    # This variable should be set if we are not providing `--answer-name` as
    # command line parameter while running yt's answer testing using nosetests.
    answer_name = None

    def __init__(self, ds_fn):
        if ds_fn is None:
            self.ds = None
        elif isinstance(ds_fn, Dataset):
            self.ds = ds_fn
        else:
            self.ds = data_dir_load(ds_fn, kwargs={"unit_system": "code"})

    def __call__(self):
        if AnswerTestingTest.result_storage is None:
            return
        nv = self.run()

        # Test answer name should be provided either as command line parameters
        # or by setting AnswerTestingTest.answer_name
        if self.options.answer_name is None and self.answer_name is None:
            raise YTNoAnswerNameSpecified()

        # This is for running answer test when `--answer-name` is not set in
        # nosetests command line arguments. In this case, set the answer_name
        # from the `answer_name` keyword in the test case
        if self.options.answer_name is None:
            pyver = f"py{sys.version_info.major}{sys.version_info.minor}"
            self.answer_name = f"{pyver}_{self.answer_name}"

            answer_store_dir = os.path.realpath(self.options.output_dir)
            ref_name = os.path.join(
                answer_store_dir, self.answer_name, self.answer_name
            )
            self.reference_storage.reference_name = ref_name
            self.reference_storage.answer_name = ref_name

            # If we are generating golden answers (passed --answer-store arg):
            # - create the answer directory for this test
            # - self.reference_storage.answer_name will be path to answer files
            if self.options.store_results:
                answer_test_dir = os.path.join(answer_store_dir, self.answer_name)
                if not os.path.isdir(answer_test_dir):
                    os.makedirs(answer_test_dir)
                self.reference_storage.reference_name = None

        if self.reference_storage.reference_name is not None:
            # Compare test generated values against the golden answer
            dd = self.reference_storage.get(self.storage_name)
            if dd is None or self.description not in dd:
                raise YTNoOldAnswer(f"{self.storage_name} : {self.description}")
            ov = dd[self.description]
            self.compare(nv, ov)
        else:
            # Store results, hence do nothing (in case of --answer-store arg)
            ov = None
        self.result_storage[self.storage_name][self.description] = nv

    @property
    def storage_name(self):
        if self.prefix != "":
            return f"{self.prefix}_{self.ds}"
        return str(self.ds)

    def compare(self, new_result, old_result):
        raise RuntimeError

    def create_plot(self, ds, plot_type, plot_field, plot_axis, plot_kwargs=None):
        # plot_type should be a string
        # plot_kwargs should be a dict
        if plot_type is None:
            raise RuntimeError("Must explicitly request a plot type")
        cls = getattr(pw, plot_type, None)
        if cls is None:
            cls = getattr(particle_plots, plot_type)
        plot = cls(*(ds, plot_axis, plot_field), **plot_kwargs)
        return plot

    @property
    def sim_center(self):
        """
        This returns the center of the domain.
        """
        return 0.5 * (self.ds.domain_right_edge + self.ds.domain_left_edge)

    @property
    def max_dens_location(self):
        """
        This is a helper function to return the location of the most dense
        point.
        """
        return self.ds.find_max(("gas", "density"))[1]

    @property
    def entire_simulation(self):
        """
        Return an unsorted array of values that cover the entire domain.
        """
        return self.ds.all_data()

    @property
    def description(self):
        obj_type = getattr(self, "obj_type", None)
        if obj_type is None:
            oname = "all"
        else:
            oname = "_".join(str(s) for s in obj_type)
        args = [self._type_name, str(self.ds), oname]
        args += [str(getattr(self, an)) for an in self._attrs]
        suffix = getattr(self, "suffix", None)
        if suffix:
            args.append(suffix)
        return "_".join(args).replace(".", "_")


class FieldValuesTest(AnswerTestingTest):
    _type_name = "FieldValues"
    _attrs = ("field",)

    def __init__(self, ds_fn, field, obj_type=None, particle_type=False, decimals=10):
        super().__init__(ds_fn)
        self.obj_type = obj_type
        self.field = field
        self.particle_type = particle_type
        self.decimals = decimals

    def run(self):
        obj = create_obj(self.ds, self.obj_type)
        field = obj._determine_fields(self.field)[0]
        fd = self.ds.field_info[field]
        if self.particle_type:
            weight_field = (field[0], "particle_ones")
        elif fd.is_sph_field:
            weight_field = (field[0], "ones")
        else:
            weight_field = ("index", "ones")
        avg = obj.quantities.weighted_average_quantity(field, weight=weight_field)
        mi, ma = obj.quantities.extrema(self.field)
        return [avg, mi, ma]

    def compare(self, new_result, old_result):
        err_msg = f"Field values for {self.field} not equal."
        if hasattr(new_result, "d"):
            new_result = new_result.d
        if hasattr(old_result, "d"):
            old_result = old_result.d
        if self.decimals is None:
            assert_equal(new_result, old_result, err_msg=err_msg, verbose=True)
        else:
            # What we do here is check if the old_result has units; if not, we
            # assume they will be the same as the units of new_result.
            if isinstance(old_result, np.ndarray) and not hasattr(
                old_result, "in_units"
            ):
                # coerce it here to the same units
                old_result = old_result * new_result[0].uq
            assert_allclose_units(
                new_result,
                old_result,
                10.0 ** (-self.decimals),
                err_msg=err_msg,
                verbose=True,
            )


class AllFieldValuesTest(AnswerTestingTest):
    _type_name = "AllFieldValues"
    _attrs = ("field",)

    def __init__(self, ds_fn, field, obj_type=None, decimals=None):
        super().__init__(ds_fn)
        self.obj_type = obj_type
        self.field = field
        self.decimals = decimals

    def run(self):
        obj = create_obj(self.ds, self.obj_type)
        return obj[self.field]

    def compare(self, new_result, old_result):
        err_msg = f"All field values for {self.field} not equal."
        if hasattr(new_result, "d"):
            new_result = new_result.d
        if hasattr(old_result, "d"):
            old_result = old_result.d
        if self.decimals is None:
            assert_equal(new_result, old_result, err_msg=err_msg, verbose=True)
        else:
            assert_rel_equal(
                new_result, old_result, self.decimals, err_msg=err_msg, verbose=True
            )


class ProjectionValuesTest(AnswerTestingTest):
    _type_name = "ProjectionValues"
    _attrs = ("field", "axis", "weight_field")

    def __init__(
        self, ds_fn, axis, field, weight_field=None, obj_type=None, decimals=10
    ):
        super().__init__(ds_fn)
        self.axis = axis
        self.field = field
        self.weight_field = weight_field
        self.obj_type = obj_type
        self.decimals = decimals

    def run(self):
        if self.obj_type is not None:
            obj = create_obj(self.ds, self.obj_type)
        else:
            obj = None
        if self.ds.domain_dimensions[self.axis] == 1:
            return None
        proj = self.ds.proj(
            self.field, self.axis, weight_field=self.weight_field, data_source=obj
        )
        return proj.field_data

    def compare(self, new_result, old_result):
        if new_result is None:
            return
        assert len(new_result) == len(old_result)
        nind, oind = None, None
        for k in new_result:
            assert k in old_result
            if oind is None:
                oind = np.array(np.isnan(old_result[k]))
            np.logical_or(oind, np.isnan(old_result[k]), oind)
            if nind is None:
                nind = np.array(np.isnan(new_result[k]))
            np.logical_or(nind, np.isnan(new_result[k]), nind)
        oind = ~oind
        nind = ~nind
        for k in new_result:
            err_msg = (
                "%s values of %s (%s weighted) projection (axis %s) not equal."
                % (k, self.field, self.weight_field, self.axis)
            )
            if k == "weight_field":
                # Our weight_field can vary between unit systems, whereas we
                # can do a unitful comparison for the other fields.  So we do
                # not do the test here.
                continue
            nres, ores = new_result[k][nind], old_result[k][oind]
            if hasattr(nres, "d"):
                nres = nres.d
            if hasattr(ores, "d"):
                ores = ores.d
            if self.decimals is None:
                assert_equal(nres, ores, err_msg=err_msg)
            else:
                assert_allclose_units(
                    nres, ores, 10.0 ** -(self.decimals), err_msg=err_msg
                )


class PixelizedProjectionValuesTest(AnswerTestingTest):
    _type_name = "PixelizedProjectionValues"
    _attrs = ("field", "axis", "weight_field")

    def __init__(self, ds_fn, axis, field, weight_field=None, obj_type=None):
        super().__init__(ds_fn)
        self.axis = axis
        self.field = field
        self.weight_field = weight_field
        self.obj_type = obj_type

    def _get_frb(self, obj):
        proj = self.ds.proj(
            self.field, self.axis, weight_field=self.weight_field, data_source=obj
        )
        frb = proj.to_frb((1.0, "unitary"), 256)
        return proj, frb

    def run(self):
        if self.obj_type is not None:
            obj = create_obj(self.ds, self.obj_type)
        else:
            obj = None
        proj, frb = self._get_frb(obj)
        frb.render(self.field)
        if self.weight_field is not None:
            frb.render(self.weight_field)
        d = frb.data
        for f in proj.field_data:
            # Sometimes f will be a tuple.
            d[f"{f}_sum"] = proj.field_data[f].sum(dtype="float64")
        return d

    def compare(self, new_result, old_result):
        assert len(new_result) == len(old_result)
        for k in new_result:
            assert k in old_result
        for k in new_result:
            # weight_field does not have units, so we do not directly compare them
            if k == "weight_field_sum":
                continue
            try:
                assert_allclose_units(new_result[k], old_result[k], 1e-10)
            except AssertionError:
                dump_images(new_result[k], old_result[k])
                raise


class PixelizedParticleProjectionValuesTest(PixelizedProjectionValuesTest):
    def _get_frb(self, obj):
        proj_plot = particle_plots.ParticleProjectionPlot(
            self.ds, self.axis, [self.field], weight_field=self.weight_field
        )
        return proj_plot.data_source, proj_plot.frb


class GridValuesTest(AnswerTestingTest):
    _type_name = "GridValues"
    _attrs = ("field",)

    def __init__(self, ds_fn, field):
        super().__init__(ds_fn)
        self.field = field

    def run(self):
        hashes = {}
        for g in self.ds.index.grids:
            hashes[g.id] = hashlib.md5(g[self.field].tobytes()).hexdigest()
            g.clear_data()
        return hashes

    def compare(self, new_result, old_result):
        assert len(new_result) == len(old_result)
        for k in new_result:
            assert k in old_result
        for k in new_result:
            if hasattr(new_result[k], "d"):
                new_result[k] = new_result[k].d
            if hasattr(old_result[k], "d"):
                old_result[k] = old_result[k].d
            assert_equal(new_result[k], old_result[k])


class VerifySimulationSameTest(AnswerTestingTest):
    _type_name = "VerifySimulationSame"
    _attrs = ()

    def __init__(self, simulation_obj):
        self.ds = simulation_obj

    def run(self):
        result = [ds.current_time for ds in self.ds]
        return result

    def compare(self, new_result, old_result):
        assert_equal(
            len(new_result),
            len(old_result),
            err_msg="Number of outputs not equal.",
            verbose=True,
        )
        for i in range(len(new_result)):
            assert_equal(
                new_result[i],
                old_result[i],
                err_msg="Output times not equal.",
                verbose=True,
            )


class GridHierarchyTest(AnswerTestingTest):
    _type_name = "GridHierarchy"
    _attrs = ()

    def run(self):
        result = {}
        result["grid_dimensions"] = self.ds.index.grid_dimensions
        result["grid_left_edges"] = self.ds.index.grid_left_edge
        result["grid_right_edges"] = self.ds.index.grid_right_edge
        result["grid_levels"] = self.ds.index.grid_levels
        result["grid_particle_count"] = self.ds.index.grid_particle_count
        return result

    def compare(self, new_result, old_result):
        for k in new_result:
            if hasattr(new_result[k], "d"):
                new_result[k] = new_result[k].d
            if hasattr(old_result[k], "d"):
                old_result[k] = old_result[k].d
            assert_equal(new_result[k], old_result[k])


class ParentageRelationshipsTest(AnswerTestingTest):
    _type_name = "ParentageRelationships"
    _attrs = ()

    def run(self):
        result = {}
        result["parents"] = []
        result["children"] = []
        for g in self.ds.index.grids:
            p = g.Parent
            if p is None:
                result["parents"].append(None)
            elif hasattr(p, "id"):
                result["parents"].append(p.id)
            else:
                result["parents"].append([pg.id for pg in p])
            result["children"].append([c.id for c in g.Children])
        return result

    def compare(self, new_result, old_result):
        for newp, oldp in zip(new_result["parents"], old_result["parents"]):
            assert newp == oldp
        for newc, oldc in zip(new_result["children"], old_result["children"]):
            assert newc == oldc


def dump_images(new_result, old_result, decimals=10):
    tmpfd, old_image = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    tmpfd, new_image = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    image_writer.write_projection(new_result, new_image)
    image_writer.write_projection(old_result, old_image)
    results = compare_images(old_image, new_image, 10 ** (-decimals))
    if results is not None:
        tempfiles = [
            line.strip() for line in results.split("\n") if line.endswith(".png")
        ]
        for fn in tempfiles:
            sys.stderr.write(f"\n[[ATTACHMENT|{fn}]]")
        sys.stderr.write("\n")


def ensure_image_comparability(a, b):
    # pad nans to the right and the bottom of two images to make them comparable
    # via matplotlib if they do not have the same shape
    if a.shape == b.shape:
        return a, b

    assert a.shape[2:] == b.shape[2:]

    warnings.warn(
        f"Images have different shapes {a.shape} and {b.shape}. "
        "Padding nans to make them comparable."
    )
    smallest_containing_shape = (
        max(a.shape[0], b.shape[0]),
        max(a.shape[1], b.shape[1]),
        *a.shape[2:],
    )
    pa = np.full(smallest_containing_shape, np.nan)
    pa[: a.shape[0], : a.shape[1], ...] = a

    pb = np.full(smallest_containing_shape, np.nan)
    pb[: b.shape[0], : b.shape[1], ...] = b

    return pa, pb


def compare_image_lists(new_result, old_result, decimals):
    fns = []
    for _ in range(2):
        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
        os.close(tmpfd)
        fns.append(tmpname)
    num_images = len(old_result)
    assert num_images > 0
    for i in range(num_images):
        expected = pickle.loads(zlib.decompress(old_result[i]))
        actual = pickle.loads(zlib.decompress(new_result[i]))
        expected_p, actual_p = ensure_image_comparability(expected, actual)

        mpimg.imsave(fns[0], expected_p)
        mpimg.imsave(fns[1], actual_p)
        results = compare_images(fns[0], fns[1], 10 ** (-decimals))
        if results is not None:
            tempfiles = [
                line.strip() for line in results.split("\n") if line.endswith(".png")
            ]
            for fn, img, padded in zip(
                tempfiles, (expected, actual), (expected_p, actual_p)
            ):
                # padded images are convenient for comparison
                # but what we really want to store and upload
                # are the actual results
                if padded.shape != img.shape:
                    mpimg.imsave(fn, img)
            if os.environ.get("JENKINS_HOME") is not None:
                for fn in tempfiles:
                    sys.stderr.write(f"\n[[ATTACHMENT|{fn}]]")
                sys.stderr.write("\n")
        assert_equal(results, None, results)
        for fn in fns:
            os.remove(fn)


class VRImageComparisonTest(AnswerTestingTest):
    _type_name = "VRImageComparison"
    _attrs = ("desc",)

    def __init__(self, scene, ds, desc, decimals):
        super().__init__(None)
        self.obj_type = ("vr",)
        self.ds = ds
        self.scene = scene
        self.desc = desc
        self.decimals = decimals

    def run(self):
        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
        os.close(tmpfd)
        self.scene.save(tmpname, sigma_clip=1.0)
        image = mpimg.imread(tmpname)
        os.remove(tmpname)
        return [zlib.compress(image.dumps())]

    def compare(self, new_result, old_result):
        compare_image_lists(new_result, old_result, self.decimals)


class PlotWindowAttributeTest(AnswerTestingTest):
    _type_name = "PlotWindowAttribute"
    _attrs = (
        "plot_type",
        "plot_field",
        "plot_axis",
        "attr_name",
        "attr_args",
        "callback_id",
    )

    def __init__(
        self,
        ds_fn: str,
        plot_field: str,
        plot_axis: str,
        attr_name: Optional[str] = None,
        attr_args: Optional[tuple] = None,
        decimals: Optional[int] = 12,
        plot_type: Optional[str] = "SlicePlot",
        callback_id: Optional[str] = "",
        callback_runners: Optional[tuple] = None,
    ):
        super().__init__(ds_fn)
        self.plot_type = plot_type
        self.plot_field = plot_field
        self.plot_axis = plot_axis
        self.plot_kwargs = {}
        self.attr_name = attr_name
        self.attr_args = attr_args
        self.decimals = decimals
        # callback_id is so that we don't have to hash the actual callbacks
        # run, but instead we call them something
        self.callback_id = callback_id
        if callback_runners is None:
            callback_runners = ()
        self.callback_runners = callback_runners

    def run(self):
        plot = self.create_plot(
            self.ds, self.plot_type, self.plot_field, self.plot_axis, self.plot_kwargs
        )
        for r in self.callback_runners:
            r(self, plot)
        if self.attr_name and self.attr_args:
            attr = getattr(plot, self.attr_name)
            attr(*self.attr_args[0], **self.attr_args[1])
        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
        os.close(tmpfd)
        plot.save(name=tmpname)
        image = mpimg.imread(tmpname)
        os.remove(tmpname)
        return [zlib.compress(image.dumps())]

    def compare(self, new_result, old_result):
        compare_image_lists(new_result, old_result, self.decimals)


class PhasePlotAttributeTest(AnswerTestingTest):
    _type_name = "PhasePlotAttribute"
    _attrs = ("plot_type", "x_field", "y_field", "z_field", "attr_name", "attr_args")

    def __init__(
        self,
        ds_fn,
        x_field,
        y_field,
        z_field,
        attr_name,
        attr_args,
        decimals,
        plot_type="PhasePlot",
    ):
        super().__init__(ds_fn)
        self.data_source = self.ds.all_data()
        self.plot_type = plot_type
        self.x_field = x_field
        self.y_field = y_field
        self.z_field = z_field
        self.plot_kwargs = {}
        self.attr_name = attr_name
        self.attr_args = attr_args
        self.decimals = decimals

    def create_plot(
        self, data_source, x_field, y_field, z_field, plot_type, plot_kwargs=None
    ):
        # plot_type should be a string
        # plot_kwargs should be a dict
        if plot_type is None:
            raise RuntimeError("Must explicitly request a plot type")
        cls = getattr(profile_plotter, plot_type, None)
        if cls is None:
            cls = getattr(particle_plots, plot_type)
        plot = cls(*(data_source, x_field, y_field, z_field), **plot_kwargs)
        return plot

    def run(self):
        plot = self.create_plot(
            self.data_source,
            self.x_field,
            self.y_field,
            self.z_field,
            self.plot_type,
            self.plot_kwargs,
        )
        attr = getattr(plot, self.attr_name)
        attr(*self.attr_args[0], **self.attr_args[1])
        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
        os.close(tmpfd)
        plot.save(name=tmpname)
        image = mpimg.imread(tmpname)
        os.remove(tmpname)
        return [zlib.compress(image.dumps())]

    def compare(self, new_result, old_result):
        compare_image_lists(new_result, old_result, self.decimals)


class GenericArrayTest(AnswerTestingTest):
    _type_name = "GenericArray"
    _attrs = ("array_func_name", "args", "kwargs")

    def __init__(self, ds_fn, array_func, args=None, kwargs=None, decimals=None):
        super().__init__(ds_fn)
        self.array_func = array_func
        self.array_func_name = array_func.__name__
        self.args = args
        self.kwargs = kwargs
        self.decimals = decimals

    def run(self):
        if self.args is None:
            args = []
        else:
            args = self.args
        if self.kwargs is None:
            kwargs = {}
        else:
            kwargs = self.kwargs
        return self.array_func(*args, **kwargs)

    def compare(self, new_result, old_result):
        if not isinstance(new_result, dict):
            new_result = {"answer": new_result}
            old_result = {"answer": old_result}

        assert_equal(
            len(new_result),
            len(old_result),
            err_msg="Number of outputs not equal.",
            verbose=True,
        )
        for k in new_result:
            if hasattr(new_result[k], "d"):
                new_result[k] = new_result[k].d
            if hasattr(old_result[k], "d"):
                old_result[k] = old_result[k].d
            if self.decimals is None:
                assert_almost_equal(new_result[k], old_result[k])
            else:
                assert_allclose_units(
                    new_result[k], old_result[k], 10 ** (-self.decimals)
                )


class GenericImageTest(AnswerTestingTest):
    _type_name = "GenericImage"
    _attrs = ("image_func_name", "args", "kwargs")

    def __init__(self, ds_fn, image_func, decimals, args=None, kwargs=None):
        super().__init__(ds_fn)
        self.image_func = image_func
        self.image_func_name = image_func.__name__
        self.args = args
        self.kwargs = kwargs
        self.decimals = decimals

    def run(self):
        if self.args is None:
            args = []
        else:
            args = self.args
        if self.kwargs is None:
            kwargs = {}
        else:
            kwargs = self.kwargs
        comp_imgs = []
        tmpdir = tempfile.mkdtemp()
        image_prefix = os.path.join(tmpdir, "test_img")
        self.image_func(image_prefix, *args, **kwargs)
        imgs = sorted(glob.glob(image_prefix + "*"))
        assert len(imgs) > 0
        for img in imgs:
            img_data = mpimg.imread(img)
            os.remove(img)
            comp_imgs.append(zlib.compress(img_data.dumps()))
        return comp_imgs

    def compare(self, new_result, old_result):
        compare_image_lists(new_result, old_result, self.decimals)


class AxialPixelizationTest(AnswerTestingTest):
    # This test is typically used once per geometry or coordinates type.
    # Feed it a dataset, and it checks that the results of basic pixelization
    # don't change.
    _type_name = "AxialPixelization"
    _attrs = ("geometry",)

    def __init__(self, ds_fn, decimals=None):
        super().__init__(ds_fn)
        self.decimals = decimals
        self.geometry = self.ds.coordinates.name

    def run(self):
        rv = {}
        ds = self.ds
        for i, axis in enumerate(ds.coordinates.axis_order):
            (bounds, center, display_center) = pw.get_window_parameters(
                axis, ds.domain_center, None, ds
            )
            slc = ds.slice(axis, center[i])
            xax = ds.coordinates.axis_name[ds.coordinates.x_axis[axis]]
            yax = ds.coordinates.axis_name[ds.coordinates.y_axis[axis]]
            pix_x = ds.coordinates.pixelize(axis, slc, ("gas", xax), bounds, (512, 512))
            pix_y = ds.coordinates.pixelize(axis, slc, ("gas", yax), bounds, (512, 512))
            # Wipe out invalid values (fillers)
            pix_x[~np.isfinite(pix_x)] = 0.0
            pix_y[~np.isfinite(pix_y)] = 0.0
            rv[f"{axis}_x"] = pix_x
            rv[f"{axis}_y"] = pix_y
        return rv

    def compare(self, new_result, old_result):
        assert_equal(
            len(new_result),
            len(old_result),
            err_msg="Number of outputs not equal.",
            verbose=True,
        )
        for k in new_result:
            if hasattr(new_result[k], "d"):
                new_result[k] = new_result[k].d
            if hasattr(old_result[k], "d"):
                old_result[k] = old_result[k].d
            if self.decimals is None:
                assert_almost_equal(new_result[k], old_result[k])
            else:
                assert_allclose_units(
                    new_result[k], old_result[k], 10 ** (-self.decimals)
                )


def requires_answer_testing():
    return skipif(
        AnswerTestingTest.result_storage is None,
        reason="answer testing storage is not properly setup",
    )


def requires_ds(ds_fn, big_data=False, file_check=False):
    condition = (big_data and not run_big_data) or not can_run_ds(ds_fn, file_check)
    return skipif(condition, reason="cannot load dataset")


def small_patch_amr(ds_fn, fields, input_center="max", input_weight=("gas", "density")):
    if not can_run_ds(ds_fn):
        return
    dso = [None, ("sphere", (input_center, (0.1, "unitary")))]
    yield GridHierarchyTest(ds_fn)
    yield ParentageRelationshipsTest(ds_fn)
    for field in fields:
        yield GridValuesTest(ds_fn, field)
        for dobj_name in dso:
            for axis in [0, 1, 2]:
                for weight_field in [None, input_weight]:
                    yield ProjectionValuesTest(
                        ds_fn, axis, field, weight_field, dobj_name
                    )
            yield FieldValuesTest(ds_fn, field, dobj_name)


def big_patch_amr(ds_fn, fields, input_center="max", input_weight=("gas", "density")):
    if not can_run_ds(ds_fn):
        return
    dso = [None, ("sphere", (input_center, (0.1, "unitary")))]
    yield GridHierarchyTest(ds_fn)
    yield ParentageRelationshipsTest(ds_fn)
    for field in fields:
        yield GridValuesTest(ds_fn, field)
        for axis in [0, 1, 2]:
            for dobj_name in dso:
                for weight_field in [None, input_weight]:
                    yield PixelizedProjectionValuesTest(
                        ds_fn, axis, field, weight_field, dobj_name
                    )


def _particle_answers(
    ds, ds_str_repr, ds_nparticles, fields, proj_test_class, center="c"
):
    if not can_run_ds(ds):
        return
    assert_equal(str(ds), ds_str_repr)
    dso = [None, ("sphere", (center, (0.1, "unitary")))]
    dd = ds.all_data()
    # this needs to explicitly be "all"
    assert_equal(dd["all", "particle_position"].shape, (ds_nparticles, 3))
    tot = sum(
        dd[ptype, "particle_position"].shape[0] for ptype in ds.particle_types_raw
    )
    assert_equal(tot, ds_nparticles)
    for dobj_name in dso:
        for field, weight_field in fields.items():
            particle_type = field[0] in ds.particle_types
            for axis in [0, 1, 2]:
                if not particle_type:
                    yield proj_test_class(ds, axis, field, weight_field, dobj_name)
            yield FieldValuesTest(ds, field, dobj_name, particle_type=particle_type)


def nbody_answer(ds, ds_str_repr, ds_nparticles, fields, center="c"):
    return _particle_answers(
        ds,
        ds_str_repr,
        ds_nparticles,
        fields,
        PixelizedParticleProjectionValuesTest,
        center=center,
    )


def sph_answer(ds, ds_str_repr, ds_nparticles, fields, center="c"):
    return _particle_answers(
        ds,
        ds_str_repr,
        ds_nparticles,
        fields,
        PixelizedProjectionValuesTest,
        center=center,
    )


def create_obj(ds, obj_type):
    # obj_type should be tuple of
    #  ( obj_name, ( args ) )
    if obj_type is None:
        return ds.all_data()
    cls = getattr(ds, obj_type[0])
    obj = cls(*obj_type[1])
    return obj



from functools import partial
from typing import Any, Callable, TypeVar

from yt.funcs import is_sequence
from yt.utilities.logger import ytLogger as mylog

from .field_info_container import FieldInfoContainer
from .field_plugin_registry import register_field_plugin

# workaround mypy not being confortable around decorator preserving signatures
# adapted from
# https://github.com/python/mypy/issues/1551#issuecomment-253978622
TFun = TypeVar("TFun", bound=Callable[..., Any])


class LocalFieldInfoContainer(FieldInfoContainer):
    def add_field(
        self, name, function, sampling_type, *, force_override=False, **kwargs
    ):
        from yt.fields.field_functions import validate_field_function

        validate_field_function(function)
        if isinstance(name, str) or not is_sequence(name):
            # the base method only accepts proper tuple field keys
            # and is only used internally, while this method is exposed to users
            # and is documented as usable with single strings as name
            if sampling_type == "particle":
                ftype = "all"
            else:
                ftype = "gas"
            name = (ftype, name)

        # Handle the case where the field has already been added.
        if not force_override and name in self:
            mylog.warning(
                "Field %s already exists. To override use `force_override=True`.",
                name,
            )

        return super().add_field(
            name, function, sampling_type, force_override=force_override, **kwargs
        )


# Empty FieldInfoContainer
local_fields = LocalFieldInfoContainer(None, [], None)

# we define two handles, essentially pointing to the same function but documented differently
# yt.add_field() is meant to be used directly, while yt.derived_field is documented
# as a decorator.
add_field = local_fields.add_field


class derived_field:
    # implement a decorator accepting keyword arguments to be passed down to add_field

    def __init__(self, **kwargs) -> None:
        self._kwargs = kwargs

    def __call__(self, f: Callable) -> Callable:
        partial(local_fields.add_field, function=f)(**self._kwargs)
        return f


@register_field_plugin
def setup_local_fields(registry, ftype="gas", slice_info=None):
    # This is easy.  We just update with the contents of the local_fields field
    # info container, and since they are not mutable in any real way, we are
    # fine.
    # Note that we actually don't care about the ftype here.
    for f in local_fields:
        registry._show_field_errors.append(f)
    registry.update(local_fields)
